/***************************************************************************
 *   (C) 2007 by Tomas Mandys <tomas.mandys at 2p dot cz>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation;             *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <libgen.h>

#define _GNU_SOURCE
#include <getopt.h>
#include <signal.h>
#include <string.h>
#include <arpa/inet.h>

#include <linux/netfilter/xt_RTPPROXY.h>

#define IPTRTPPROXY_VERSION "0.3"

static int verbose = 0;
static enum {fmtDefault=0, fmtXml, fmtTab} output_format = fmtDefault;
const char* program_name;

#define CMDVAL_ADDR_PORT 0x0001
#define CMDVAL_A_B 0x0002
#define CMDVAL_RTP_RTCP 0x0004

#define CMDVAL_GRP_MASK 0x1F00
#define CMDVAL_GRP_ADDR 0x0100
#define CMDVAL_GRP_RADDR 0x0200
#define CMDVAL_GRP_LEARNING_TIMEOUT 0x0300
#define CMDVAL_GRP_ALWAYS_LEARN 0x0400
#define CMDVAL_GRP_SESSION_ID 0x0500
#define CMDVAL_GRP_LIST 0x0600
#define CMDVAL_GRP_RESET 0x0700
#define CMDVAL_GRP_SESSION_CREATED 0x0800
#define CMDVAL_GRP_FORMAT 0x1000

static struct option long_options[] = {
	{ "addr-a",			1, 0, CMDVAL_GRP_ADDR },
	{ "addr-b",			1, 0, CMDVAL_GRP_ADDR|CMDVAL_A_B },
	{ "port-a",			1, 0, CMDVAL_GRP_ADDR|CMDVAL_ADDR_PORT },
	{ "port-b",			1, 0, CMDVAL_GRP_ADDR|CMDVAL_ADDR_PORT|CMDVAL_A_B },
	{ "rtp-addr-a",			1, 0, CMDVAL_GRP_RADDR },
	{ "rtcp-addr-a",		1, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP },
	{ "rtp-port-a",			1, 0, CMDVAL_GRP_RADDR|CMDVAL_ADDR_PORT },
	{ "rtcp-port-a",		1, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP|CMDVAL_ADDR_PORT },
	{ "rtp-addr-b",			1, 0, CMDVAL_GRP_RADDR|CMDVAL_A_B },
	{ "rtcp-addr-b",		1, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP|CMDVAL_A_B },
	{ "rtp-port-b",			1, 0, CMDVAL_GRP_RADDR|CMDVAL_ADDR_PORT|CMDVAL_A_B },
	{ "rtcp-port-b",		1, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP|CMDVAL_ADDR_PORT|CMDVAL_A_B },
	{ "rtp-learning-timeout-a",	1, 0, CMDVAL_GRP_LEARNING_TIMEOUT },
	{ "rtp-learning-timeout-b",	1, 0, CMDVAL_GRP_LEARNING_TIMEOUT|CMDVAL_A_B },
	{ "rtcp-learning-timeout-a",	1, 0, CMDVAL_GRP_LEARNING_TIMEOUT|CMDVAL_RTP_RTCP },
	{ "rtcp-learning-timeout-b",	1, 0, CMDVAL_GRP_LEARNING_TIMEOUT|CMDVAL_RTP_RTCP|CMDVAL_A_B },
	{ "always-learn-a",		0, 0, CMDVAL_GRP_ALWAYS_LEARN },
	{ "always-learn-b",		0, 0, CMDVAL_GRP_ALWAYS_LEARN|CMDVAL_A_B },

	{ "sess-id-lo",			1, 0, CMDVAL_GRP_SESSION_ID },
	{ "sess-id-hi",			1, 0, CMDVAL_GRP_SESSION_ID|1 },
	{ "sess-id-created",            1, 0, CMDVAL_GRP_SESSION_CREATED},

	{ "no-switchboard",		0, 0, CMDVAL_GRP_LIST|1 },
	{ "no-session",			0, 0, CMDVAL_GRP_LIST|2 },

	{ "reset-global-stat",		0, 0, CMDVAL_GRP_RESET|0 },
	{ "reset-switchboard-stat",	0, 0, CMDVAL_GRP_RESET|1 },
	{ "reset-packet-stat",		0, 0, CMDVAL_GRP_RESET|2 },
	{ "force-switchboard-audit",	0, 0, CMDVAL_GRP_RESET|3 },

	{ "format",			1, 0, CMDVAL_GRP_FORMAT },
	{ "verbose",			0, 0, 'v' },
	{ "help",			0, 0, 'h' },
	{ 0 }
};

struct enum_entry {
        int id;
        char* name;
        char* descr;
};

enum {cmdNone=0, cmdInfo, cmdList, cmdAlloc, cmdUpdate, cmdDelete};

static struct enum_entry enum_command[] = {
	{id: cmdInfo, 	name: "info", descr: 
		"get RTP proxy global information"
	},
	{id: cmdList, 	name: "list",	descr: 
		"list RTP sessions of specified switchboard. If no switchboard specified then all switchboards are listed. "
		"Session-id may specify session range. "
		"Specify verbosity via --list-xxx options."
	},
	{id: cmdAlloc,	name: "alloc",	descr: 
		"alloc new RTP session. If switchboard is not specified then session is allocated at any switchboard. "
		"At least one remote client addr/port specification is mandatory."
	},
	{id: cmdUpdate,	name: "update",	descr: 
		"update RTP session(s) at specified switchboard. If session-id range is not specified then all sessions of particular switchboard are updated."
	},
	{id: cmdDelete,	name: "delete",	descr: 
		"delete RTP session(s). If switchboard is not specified then all switchboard sessions are deleted according session-id. "
		"If session-id is not specified then all sessions are deleted."
	},

	{id: cmdNone, name: NULL}
};

static struct enum_entry enum_format_option[] = {
	{id: fmtDefault, name: "default", descr:
		"human readable (default)"
	},
	{id: fmtXml, name: "xml", descr:
		"XML"
	},
	{id: fmtTab, name: "tab", descr:
		"tab-delimited"
	},
	{id: -1, name: NULL}
};

static struct enum_entry *enum_by_id(int id, struct enum_entry *e) {
	int i;
	for (i=0; e[i].name != NULL; i++) {
		if (e[i].id == id) 
			return e+i;
	}
	return NULL;
}

static struct enum_entry *enum_by_name(char *name, struct enum_entry *e) {
	int i;
	for (i=0; e[i].name != NULL; i++) {
		if (strcmp(name, e[i].name)==0) 
			return e+i;
	}
	return NULL;
}

static void print_opt (int cmd) {
	struct option *o;
	int i;
	for (o=long_options; o->name; o++) {
		if ((o->val & CMDVAL_GRP_MASK) == cmd) {
			printf("  --%s ", o->name);
			switch (cmd) {
				case CMDVAL_GRP_ADDR:
					printf("<%s  RTP proxy gate-%c %s", 
						(o->val & CMDVAL_ADDR_PORT)?"port>":"ip>  ",
						'a'+((o->val & CMDVAL_A_B) != 0),
						(o->val & CMDVAL_ADDR_PORT)?"port":"ip"
					);
					break;

				case CMDVAL_GRP_RADDR:
					printf("<%s  %s client-%c %s", 
						(o->val & CMDVAL_ADDR_PORT)?"port>":"ip>  ",
						(o->val & CMDVAL_RTP_RTCP)?"RTCP":" RTP",
						'a'+((o->val & CMDVAL_A_B) != 0),
						(o->val & CMDVAL_ADDR_PORT)?"port":"ip"
					);
					break;

				case CMDVAL_GRP_LEARNING_TIMEOUT:
					printf("<msec>  %s client-%c learning timeout", 
						(o->val & CMDVAL_RTP_RTCP)?"RTCP":" RTP",
						'a'+((o->val & CMDVAL_A_B) != 0)
					);
					break;

				case CMDVAL_GRP_ALWAYS_LEARN:
					printf("     always learn RTP client-%c", 
						'a'+((o->val & CMDVAL_A_B) != 0)
					);
					break;

				case CMDVAL_GRP_SESSION_ID:
					printf("<id>  session-id %s limit", 
						(o->val & 0x0F)?"hi":"lo"
					);
					break;

				case CMDVAL_GRP_SESSION_CREATED:
					printf("<tick>  additional session identification key (optional)");

				case CMDVAL_GRP_LIST:
					for (i=strlen(o->name); i<16; i++) printf(" ");
					switch (o->val & 0x0F) {
						case 1:
							printf("  don't list switchboards and sessions");
							break; 
						case 2:
							printf("  don't list sessions");
							break; 
						default: ;
					}
					break;

				case CMDVAL_GRP_RESET:
					for (i=strlen(o->name); i<23; i++) printf(" ");
					switch (o->val & 0x0F) {
						case 0:
							printf("  reset global statistics");
							break; 
						case 1:
							printf("  reset switchboard statistics");
							break; 
						case 2:
							printf("  reset switchboard packet statistics");
							break; 
						case 3:
							printf("  force switchboard audit");
							break; 
						default: ;
					}
					break;

				case CMDVAL_GRP_FORMAT:
					printf("<fmt>      output format");
					for (i=0; enum_format_option[i].name; i++) {
						printf("\n            %s ..  %s", enum_format_option[i].name,  enum_format_option[i].descr);
					}
					break;	

				default: ;
			}
			printf("\n");
		}
	}
}

static void print_version() {
	printf("iptables RTPPROXY helper, v%s\n", IPTRTPPROXY_VERSION);
}

static void help(int cmd) {
	struct enum_entry *e;
	print_version();
	if (!cmd) {
		printf("Usage: %s [options] <command> [options]\n", program_name);
		printf("Type '%s -h <command>' for help on a specific command.\n", program_name);

		printf("\n");
		printf("Command list:\n");
		for (e=enum_command; e->name; e++) {
			printf("  %-8s  %s\n", e->name, e->descr);
		}
	}
	else {
		e = enum_by_id(cmd, enum_command);
		printf("Usage: %s %s [options]\n", program_name, e->name);

		printf("%s: %s\n", e->name, e->descr);

		if (cmd != cmdInfo) {
			printf( "\n"
				"Switchboard identification:\n");
			print_opt(CMDVAL_GRP_ADDR);
	
			if (cmd != cmdAlloc) {
				printf( "\n"
					"Session range:\n");
				print_opt(CMDVAL_GRP_SESSION_ID);
				print_opt(CMDVAL_GRP_SESSION_CREATED);
			}
	
			if (cmd == cmdAlloc || cmd == cmdUpdate) {
				printf( "\n"
					"RTP client parameters:\n");
				print_opt(CMDVAL_GRP_RADDR);
				print_opt(CMDVAL_GRP_LEARNING_TIMEOUT);
				print_opt(CMDVAL_GRP_ALWAYS_LEARN);
			}

			if (cmd == cmdList) {
				printf( "\n"
					"List verbosity:\n");
				print_opt(CMDVAL_GRP_LIST);
			}
		}
		printf( "\n"
			"Other options:\n");
		print_opt(CMDVAL_GRP_RESET);
		if (cmd == cmdList || cmd == cmdAlloc) {
			print_opt(CMDVAL_GRP_FORMAT);
		}
	}

	printf("\n");
	printf("Common options:\n");
	printf("  -h,--help      print this help\n");
	printf("  -v,--verbose   print more info, mostly to stderr\n");
	printf("  -V             print version\n");
	printf("\n");
}

static char *ip2s(u_int32_t ip) {
	struct in_addr ip2 = { ip };
	return inet_ntoa(ip2);
} 
/*static char *ip2s(u_int32_t addr) {
    static char buf[17];
    const unsigned char *bytep;

    bytep = (const unsigned char *) &addr;
    snprintf(buf, 16, "%u.%u.%u.%u", bytep[0], bytep[1], bytep[2], bytep[3]);
    buf[16] = 0;
    return buf;
}
*/

static void print_header(struct xt_rtpproxy_handle *handle) {
	switch (output_format) {
		case fmtXml:
			printf("<header version=\"0x%x\" flags=\"0x%x\" />\n", handle->info.h.version, handle->info.h.flags);
			break;
		default:
			printf("header: {version: 0x%x, flags: 0x%x}\n", handle->info.h.version, handle->info.h.flags);
			break;
	}
}

static void print_stamp(u_int32_t t) {
	time_t tt;
	char *c, *c2;	
	tt = (time_t) t;
	c = ctime(&tt);
	for (c2 = c+strlen(c); c2 >= c && *c2 <= ' ' ; c2--) {
		*c2 = '\0';
	}
	switch (output_format) {

		case fmtXml:
			printf("tick=\"%u\" time=\"%s\"", t, c);
			break;
		case fmtTab:
			printf("%u\t%s", t, c);
			break;

		default:
			printf("{tick: %u, time: %s}", t, c);
			break;
	};
};

static void print_info(struct xt_rtpproxy_handle *handle) {
	static char *cmd_s[XT_RTPPROXY_MAX_CMD] = {"update", "delete", "config", "alloc"};
	int i;
	if (verbose && output_format!=fmtTab) print_header(handle);

	switch (output_format) {

		case fmtXml:
			printf("<started "); print_stamp(handle->info.started); printf(" />\n");
			printf("<total switchboards=\"%u\" sessions=\"%u\" />\n", handle->info.total.num_switchboard, handle->info.total.num_session); 
			printf("<stat ok=\"%u\" fail=\"%u\" perm=\"%u\" unknown=\"%u\"", 
				handle->info.stat.ok, 
				handle->info.stat.fail, 
				handle->info.stat.perm, 
				handle->info.stat.unknown
			);
			for (i=0; i<XT_RTPPROXY_MAX_CMD; i++) {
				printf(" %s=\"%u\"", cmd_s[i], handle->info.stat.cmd[i]); 
			}
			printf(" />\n");
			break;
		case fmtTab:
			printf("started/tick\tstarted/datetime");
			printf("\ttotal/sessions\ttotal/sessions\tstat/ok\tstat/fail\tstat/perm\tstat/unknown");
			for (i=0; i<XT_RTPPROXY_MAX_CMD; i++) {
				printf("\tstat/%s", cmd_s[i]); 
			}
			printf("\n");
			print_stamp(handle->info.started);
			printf("\t%u\t%u", handle->info.total.num_switchboard, handle->info.total.num_session);
			printf("\t%u\t%u\t%u\t%u", 
				handle->info.stat.ok, 
				handle->info.stat.fail, 
				handle->info.stat.perm, 
				handle->info.stat.unknown
			);
			for (i=0; i<XT_RTPPROXY_MAX_CMD; i++) {
				printf("\t%u", handle->info.stat.cmd[i]); 
			}
			printf("\n");
			break;

		default:
			printf("started: "); print_stamp(handle->info.started); printf("\n");
			printf("total: {switchboards: %u, sessions: %u}\n", handle->info.total.num_switchboard, handle->info.total.num_session); 
			printf("stat: {ok: %u, fail: %u, perm: %u, unknown: %u", 
				handle->info.stat.ok, 
				handle->info.stat.fail, 
				handle->info.stat.perm, 
				handle->info.stat.unknown
			);
			for (i=0; i<XT_RTPPROXY_MAX_CMD; i++) {
				printf(", %s: %u", cmd_s[i], handle->info.stat.cmd[i]); 
			}
			printf("}\n");
			break;
	}
}

static void print_stat_title(char *gate) {
	switch (output_format) {
		case fmtTab:
			printf("%s/stat/packets/in\t%s/stat/packets/out\t%s/stat/bytes/in\t%s/stat/bytes/out\t", 
				gate, gate, gate, gate);
			break;
		default:;
	}
}

static void print_stat(struct xt_rtpproxy_stat *stat) {
	switch (output_format) {
		case fmtXml:
			printf("<stat><packets in=\"%u\" out=\"%u\" /><bytes in=\"%u\" out=\"%u\" /></stat>\n",
				stat->packets_in,
				stat->packets_out,
				stat->bytes_in,
				stat->bytes_out
			);
			break;
		case fmtTab:
			printf("%u\t%u\t%u\t%u",
				stat->packets_in,
				stat->packets_out,
				stat->bytes_in,
				stat->bytes_out
			);
			break;
		default:
			printf("packets: {in: %u, out: %u}, bytes: {in: %u, out: %u}",
				stat->packets_in,
				stat->packets_out,
				stat->bytes_in,
				stat->bytes_out
			);
	};
};

static void print_in_switchboard(struct xt_rtpproxy_sockopt_in_switchboard *s) {
	fprintf(stderr, "  switchboard: ");
	if (s) {
		fprintf(stderr, "{gate-a: {ip: %s, port: %d}, ", 
			ip2s(s->gate[0].ip), 
			s->gate[0].port
		); /* ip2s ... result in static buffer */
		fprintf(stderr, "gate-b: {ip: %s, port: %d}}", 
			ip2s(s->gate[1].ip), 
			s->gate[1].port
		); /* ip2s ... result in static buffer */
	}
	else {
		fprintf(stderr, "NULL");
	}
}

static void print_in_session(struct xt_rtpproxy_sockopt_in_alloc_session *s) {
	int i, j;
	fprintf(stderr, "  session: ");
	if (s) {
		fprintf(stderr, "{\n");
		for (i=0; i<XT_RTPPROXY_MAX_GATE; i++) {
			fprintf(stderr, "    gate-%c: {\n", 'a'+i);
			for (j=0; j<2; j++) {
				if (s->source[i].stream[j].flags != 0) {
					fprintf(stderr, "      %s: {", j?"rtcp":"rtp");
					if (s->source[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR) {
						fprintf(stderr, "addr: {ip: %s, port: %d}", ip2s(s->source[i].stream[j].addr.ip), s->source[i].stream[j].addr.port);
					}
					if (s->source[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT) {
						if (s->source[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR)
							fprintf(stderr, ", ");
						fprintf(stderr, "learn-timeout: %d", s->source[i].stream[j].learning_timeout);
					}
					fprintf(stderr, "},\n");
				}
			}
			fprintf(stderr, "      always-learn: %d\n", s->source[i].always_learn);
			fprintf(stderr, "    }%s\n", (i<XT_RTPPROXY_MAX_GATE-1)?",":"");
		}
		fprintf(stderr, "  }");
	}
	else {
		fprintf(stderr, "NULL");
	}
}

static void print_in_session_id(struct xt_rtpproxy_sockopt_in_sess_id *s) {
	fprintf(stderr, "  sess-id: ");
	if (s) {
		fprintf(stderr, "{min: %d, max: %d, created: %u}", s->sess_id_min, s->sess_id_max, s->created);
	}
	else {
		fprintf(stderr, "NULL");
	}
}

static void print_format() {
	struct enum_entry *e = enum_by_id(output_format, enum_format_option);
	if (e) fprintf(stderr, "format: %s\n", e->name); 
}

static char stop_flag = 0;
static void sig_term(int signr) {
    signal(SIGINT, SIG_IGN);
    signal(SIGQUIT, SIG_IGN);
    signal(SIGTERM, SIG_IGN);
    stop_flag = 1;
}
#define SIGNAL(sig) \
	if (signal(sig, sig_term) == SIG_ERR) {\
		printf("can't install signal handler for %s\n", #sig);\
	exit (-1);\
}
 
static u_int32_t s2ip(char *optarg) {
	u_int32_t res = 0;
	if (!inet_aton(optarg, (struct in_addr*) &res)) {
		fprintf(stderr, "Invalid IP address %s\n", optarg);
		exit(-1);
	}
	return res;
}

static u_int32_t s2u(char *optarg) {
	long res = atol(optarg);
	if (res < 0) {
		fprintf(stderr, "Invalid number %s\n", optarg);
		exit(-1);
	}
	return res;
}

int main(int argc, char** argv) { 

	int cmd = cmdNone, help_flag = 0;
	static char short_opts[100] = "+HvV";
	struct xt_rtpproxy_sockopt_in_alloc_session in_session;
	struct xt_rtpproxy_sockopt_in_switchboard in_switchboard;
	struct xt_rtpproxy_sockopt_in_sess_id in_sess_id;
	struct xt_rtpproxy_handle handle;
	int i, j, in_switchboard_fl, in_sess_id_fl;
	u_int32_t in_flags, list_verbose;

	program_name = basename(argv[0]);
	if (argc <= 1) {
		help(cmd);
		return 2;
	}

	memset(&in_session, 0, sizeof(in_session));
	memset(&in_switchboard, 0, sizeof(in_switchboard));
	memset(&in_sess_id, 0xFF, sizeof(in_sess_id));
	in_sess_id.created = 0;

	j = strlen(short_opts);
	for (i=0; long_options[i].name; i++) {
		if (long_options[i].val > ' ' && long_options[i].val < 128) {
			short_opts[j++] = (char) long_options[i].val;
			if (long_options[i].has_arg != no_argument) 
				short_opts[j++] = ':';
			if (long_options[i].has_arg == optional_argument) 
				short_opts[j++] = ':';
			short_opts[j] = '\0';
		}
	}
	in_flags = 0;
	list_verbose = 0;
	while (optind < argc) {
		int option_index;
		int c;
		option_index = -1;
		c = getopt_long_only (argc, argv, short_opts, long_options, &option_index);
		if (c == -1) {
			if (optind >= argc) 
				break;

			if (!cmd) {
				struct enum_entry *e;
				e = enum_by_name(argv[optind], enum_command);
				if (!e) {
					fprintf(stderr, "ERROR: Unknown command: %s\n", argv[optind]);
					goto destroy;
				}
				cmd = e->id;
				optind++;
			} 
			else {
				fprintf(stderr, "ERROR: Multiple commands not allowed (%s)\n", argv[optind]);
				return 2;
			}
		}
		else {
			
			switch ((c & CMDVAL_GRP_MASK)?c & CMDVAL_GRP_MASK:c) {
				case CMDVAL_GRP_ADDR:
					if ((c & CMDVAL_ADDR_PORT) != 0)
						in_switchboard.gate[(c & CMDVAL_A_B) != 0].port = s2u(optarg);
					else
						in_switchboard.gate[(c & CMDVAL_A_B) != 0].ip = s2ip(optarg);
					break;
				case CMDVAL_GRP_RADDR:
					if ((c & CMDVAL_ADDR_PORT) != 0)
						in_session.source[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].addr.port = s2u(optarg);
					else
						in_session.source[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].addr.ip = s2ip(optarg);
					in_session.source[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].flags |= XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR;
					break;
				case CMDVAL_GRP_LEARNING_TIMEOUT:
					in_session.source[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].learning_timeout = s2u(optarg);
					in_session.source[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].flags |= XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT;
					break;
				case CMDVAL_GRP_ALWAYS_LEARN:
					in_session.source[(c & CMDVAL_A_B) != 0].always_learn = 1;
					break;
				case CMDVAL_GRP_SESSION_ID:
					if (c & 0x01)
						in_sess_id.sess_id_max = s2u(optarg);
					else
						in_sess_id.sess_id_min = s2u(optarg);
					break;
				case CMDVAL_GRP_SESSION_CREATED:
					in_sess_id.created = s2u(optarg);
					break;
				case CMDVAL_GRP_LIST:
					switch (c & 0x0F) {
						case 1:
							list_verbose |= XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD | XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION;
							break;
						case 2:
							list_verbose |= XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION;
							break;
						default: ;
					}
					break;

				case CMDVAL_GRP_RESET:
					switch (c & 0x0F) {
						case 0:
							in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_RESET_GLOBAL_STAT;
							break;
						case 1:
							in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_RESET_SWITCHBOARD_STAT;
							break;
						case 2:
							in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_RESET_SWITCHBOARD_PACKET_STAT;
							break;
						case 3:
							in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_FORCE_SWITCHBOARD_AUDIT;
							break;
						default: ;
					}
					break;
				case CMDVAL_GRP_FORMAT: {
					struct enum_entry *e;
					e = enum_by_name(optarg, enum_format_option);
					if (!e) {
						fprintf(stderr, "ERROR: Unknown format option: %s\n", optarg);
						goto destroy;
					}
					output_format = e->id;
					break;
					}
				case 'v':
					verbose++;
					break;
				case 'V':
					print_version();
					return 0;
				case 'h':
				case 'H':
					help_flag = 1;
					break;
				case '?':
					break;
				default:
					fprintf(stderr, "getopt_long_only returned character code 0x%x\n", c);
					break;
			}
		}
		if (cmd && help_flag)
			break;
	}
	if (!cmd || help_flag) {
		help(cmd);
		return help_flag?0:2;
	}

	SIGNAL(SIGTERM);
	SIGNAL(SIGINT);
	SIGNAL(SIGQUIT);

	/* switchboard may be optional but when at least one port or ip is defined then complete definition is mandatory */
	in_switchboard_fl = 0;
	for (i=0; i<XT_RTPPROXY_MAX_GATE; i++) {
		if (in_switchboard.gate[i].ip || in_switchboard.gate[i].port)
			in_switchboard_fl = 1;
		if (i>0) {
			if (!in_switchboard.gate[i].ip)
				in_switchboard.gate[i].ip = in_switchboard.gate[0].ip;
			if (!in_switchboard.gate[i].port)
				in_switchboard.gate[i].port = in_switchboard.gate[0].port;
		}
		if (in_switchboard_fl && (!in_switchboard.gate[i].ip || !in_switchboard.gate[i].port)) {
			fprintf(stderr, "ERROR: uncomplete switchboard identification\n");
			return 2;
		} 
	}

	if (in_sess_id.sess_id_max == 0xFFFF)
		in_sess_id.sess_id_max = in_sess_id.sess_id_min;
	if (in_sess_id.sess_id_min == 0xFFFF)
		in_sess_id.sess_id_min = in_sess_id.sess_id_max;
	if (in_sess_id.sess_id_max < in_sess_id.sess_id_min) {
		u_int16_t x = in_sess_id.sess_id_max;
		in_sess_id.sess_id_max = in_sess_id.sess_id_min;
		in_sess_id.sess_id_min = x;
	}
	in_sess_id_fl = in_sess_id.sess_id_min != 0xFFFF && in_sess_id.sess_id_max != 0xFFFF;
	
	if (verbose) fprintf(stderr, "xt_RTPPROXY_open()\n");
	if (xt_RTPPROXY_open(&handle) < 0) goto err;
	
	if (output_format == fmtXml) printf("<iptrtpproxy>\n");
	switch (cmd) {
		case cmdInfo:
			if (verbose) fprintf(stderr, "xt_RTPPROXY_get_info(flags:0x%x)\n", in_flags);
			if (xt_RTPPROXY_get_info(&handle, in_flags) < 0)
				goto err;
			if (verbose) print_format();
			print_info(&handle);
			break;

		case cmdList: {
			struct xt_rtpproxy_switchboard *out_switchboard, *sb;
			static char* state_s[] = {"none", "init1", "init2", "forward1", "forward2", "destroyed", "expired"};
			in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD | XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION;
			in_flags &= ~list_verbose;  
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_get_switchboards(\n");
				print_in_switchboard(in_switchboard_fl?&in_switchboard:NULL); fprintf(stderr, ",\n");
				print_in_session_id(in_sess_id_fl?&in_sess_id:NULL); fprintf(stderr, ",\n");
				fprintf(stderr, "  flags: 0x%x\n", in_flags); 
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPROXY_get_switchboards(&handle, in_switchboard_fl?&in_switchboard:NULL, in_sess_id_fl?&in_sess_id:NULL, in_flags, &out_switchboard) < 0)
				goto err;
			if (verbose) print_format(); 
			if (verbose && output_format!=fmtTab) print_info(&handle);

			if (in_flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD) {
				switch (output_format) {
					case fmtXml:
						printf("<switchboards>\n");
						for (sb=out_switchboard; sb; sb=sb->next) {
							printf("  <switchboard max-sess=\"%d\" expires-timeout=\"%d\" resurrection-timeout=\"%d\">\n", 
								sb->so.max_sess, 
								sb->so.expires_timeout, 
								sb->so.resurrection_timeout
							);

							printf("    <gate id=\"a\" ip=\"%s\" port=\"%d\">\n", 
		                				ip2s(sb->so.gate[0].addr.ip),
						                sb->so.gate[0].addr.port
						        ); /* ip2s ... result in static buffer */
							printf("      "); print_stat(&sb->so.gate[0].stat);
							printf("    </gate>\n");
							printf("    <gate id=\"b\" ip=\"%s\" port=\"%d\">\n", 
                						ip2s(sb->so.gate[1].addr.ip),
				                		sb->so.gate[1].addr.port
						        ); /* ip2s ... result in static buffer */
							printf("      "); print_stat(&sb->so.gate[1].stat);
							printf("    </gate>\n");
							printf("    <stat>\n");
							printf("      <state");
							for (j=0; j<xt_rtpproxy_MAX_STATES; j++) {
								printf(" %s=\"%u\"", state_s[j], sb->so.stat.state[j]);
							}
							printf(" />\n");
							printf("      <counter alloc=\"%u\" alloc-failed=\"%u\" updated=\"%u\" destroyed=\"%u\" expired=\"%u\" />\n",
								sb->so.stat.counter.alloc,
								sb->so.stat.counter.alloc_failed,
								sb->so.stat.counter.updated,
								sb->so.stat.counter.destroyed,
								sb->so.stat.counter.expired 
							);
							printf("    </stat>\n");
							if (in_flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION) {
								struct xt_rtpproxy_session *ss;
								printf("    <sessions>\n");
								for (ss=sb->sessions; ss; ss=ss->next) {
									int g;
									printf("      <session state=\"%s\" sess-id=\"%d\">", state_s[ss->state], ss->sess_id);
									printf("<created "); print_stamp(ss->created); printf(" />");
									printf("<expires "); print_stamp(ss->expires); printf(" />");
									printf("\n");
									for (g=0; g<XT_RTPPROXY_MAX_GATE; g++) {
										int st;
										printf("        <gate id=\"%c\" always-learn=\"%d\">\n", 'a'+g, ss->gate[g].always_learn);
										for (st=0; st<2; st++) {
											printf("          <port proto=\"%s\" value=\"%d\"><source ip=\"%s\" port=\"%d\" />\n",
												st==0?"rtp":"rtcp",
												ss->gate[g].stream[st].port,
												ip2s(ss->gate[g].stream[st].source.ip),
												ss->gate[g].stream[st].source.port
											);
											printf("            <learning ");
											print_stamp(ss->gate[g].stream[st].learning),
											printf(" /> <last_packet ");
											print_stamp(ss->gate[g].stream[st].last_packet_stamp);
											printf("/>\n");
											printf("          </port>\n");
										}
										printf("          "); print_stat(&ss->gate[g].stat); 
										printf("        </gate>\n");
									}
									printf("      </session>\n");
								}
								printf("    </sessions>\n");
							}
							printf("  </switchboard>\n");
						}
						printf("</switchboards>\n");
						break;

					case fmtTab:
						if ((in_flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION) == 0) {
							printf("gate-a/ip\tgate-a/port\tgate-b/ip\tgate-b/port\tmax-sess\texpires-timeout\tresurrection-timeout\t");
							print_stat_title("gate-a");
							print_stat_title("gate-b");
							for (j=0; j<xt_rtpproxy_MAX_STATES; j++) {
								printf("stat/%s\t", state_s[j]);
							}
							printf("counter/alloc\tcounter/alloc-failed\tcounter/updated\tcounter/destroyed\tcounter/expired");
							printf("\n");

							for (sb=out_switchboard; sb; sb=sb->next) {
								printf("%s\t%d\t", 
									ip2s(sb->so.gate[0].addr.ip),
									sb->so.gate[0].addr.port
								); /* ip2s ... result in static buffer */
								printf("%s\t%d\t", 
									ip2s(sb->so.gate[1].addr.ip),
									sb->so.gate[1].addr.port
								); /* ip2s ... result in static buffer */
								printf("%d\t%d\t%d\t", 
									sb->so.max_sess, 
									sb->so.expires_timeout, 
									sb->so.resurrection_timeout
								);
								print_stat(&sb->so.gate[0].stat); printf("\t");
								print_stat(&sb->so.gate[1].stat); printf("\t");

								for (j=0; j<xt_rtpproxy_MAX_STATES; j++) {
									printf("%u\t", sb->so.stat.state[j]);
								}
								printf("%u\t%u\t%u\t%u\t%u",
									sb->so.stat.counter.alloc,
									sb->so.stat.counter.alloc_failed,
									sb->so.stat.counter.updated,
									sb->so.stat.counter.destroyed,
									sb->so.stat.counter.expired 
								);
								printf("\n");
							}
						}
						else {
							int g;
							printf("gate-a/ip\tgate-a/port\tgate-b/ip\tgate-b/port\tsess-id\tcreated/tick\tcreated/datetime\texpires/tick\texpires/datetime\t");

							for (g=0; g<XT_RTPPROXY_MAX_GATE; g++) {
								int st;
								printf("gate-%c/always-learn\t", 'a'+g);
								for (st=0; st<2; st++) {
									char c[15];
									sprintf(c, "gate-%c/%s", 'a'+g, st?"rtcp":"rtp");
									printf("%s/port\t%s/source/ip\t%s/source/port\t"
										"%s/learning/tick\t%s/learning/datetime\t"
										"%s/last-packet/tick\t%s/last-packet/datetime\t",
										c, c, c, c, c, c, c);
									print_stat_title(c);

								}
							}
							printf("\n");

							for (sb=out_switchboard; sb; sb=sb->next) {
								struct xt_rtpproxy_session *ss;
								for (ss=sb->sessions; ss; ss=ss->next) {
									printf("%s\t%d\t", 
										ip2s(sb->so.gate[0].addr.ip),
										sb->so.gate[0].addr.port
									); /* ip2s ... result in static buffer */
									printf("%s\t%d\t", 
										ip2s(sb->so.gate[1].addr.ip),
										sb->so.gate[1].addr.port
									); /* ip2s ... result in static buffer */
									printf("%d\t", 
										ss->sess_id 
									);
									print_stamp(ss->created); printf("\t");
									print_stamp(ss->expires); printf("\t");
									for (g=0; g<XT_RTPPROXY_MAX_GATE; g++) {
										int st;
										printf("%d", ss->gate[g].always_learn);
										for (st=0; st<2; st++) {
											printf("\t%d\t%s\t%d\t",
												ss->gate[g].stream[st].port,
												ip2s(ss->gate[g].stream[st].source.ip),
												ss->gate[g].stream[st].source.port
											);
											print_stamp(ss->gate[g].stream[st].learning); printf("\t");
											print_stamp(ss->gate[g].stream[st].last_packet_stamp); printf("\t");
											print_stat(&ss->gate[g].stat); 
										}
									}
									printf("\n");
								}
							}
						}
						break;

					default:
						printf("{\n");
						for (sb=out_switchboard; sb; sb=sb->next) {
							int j;
							printf("  {\n");
							printf("    switchboard: {\n");
							printf("      gate: {\n");
							printf("        a: {ip: %s, port: %d}, stat: ",
								ip2s(sb->so.gate[0].addr.ip), 
								sb->so.gate[0].addr.port
							); /* ip2s ... result in static buffer */
							print_stat(&sb->so.gate[0].stat);
							printf("\n");
							printf("        b: {ip: %s, port: %d}, stat: ", 
								ip2s(sb->so.gate[1].addr.ip),
								sb->so.gate[1].addr.port
							); /* ip2s ... result in static buffer */
							print_stat(&sb->so.gate[1].stat);
							printf("\n      }\n");
							printf("      max-sess: %d, expires-timeout: %d, resurrection-timeout: %d\n", 
								sb->so.max_sess, 
								sb->so.expires_timeout, 
								sb->so.resurrection_timeout
							);
							printf("      stat: {\n");
							printf("        state: {");
							for (j=0; j<xt_rtpproxy_MAX_STATES; j++) {
								if (j) printf(", ");
								printf("%s: %u", state_s[j], sb->so.stat.state[j]);
							}
							printf("}\n");
							printf("        counter: {alloc: %u, alloc-failed: %u, updated: %u, destroyed: %u, expired: %u}\n",
								sb->so.stat.counter.alloc,
								sb->so.stat.counter.alloc_failed,
								sb->so.stat.counter.updated,
								sb->so.stat.counter.destroyed,
								sb->so.stat.counter.expired 
							);
							printf("      }\n");
							printf("    }\n");
							if (in_flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION) {
								struct xt_rtpproxy_session *ss;
								printf("    sessions: {\n");
								for (ss=sb->sessions; ss; ss=ss->next) {
									int g;
									printf("      {\n");
									printf("        state: %s, sess-id: %d, created: ", state_s[ss->state], ss->sess_id);
									print_stamp(ss->created);
									printf(", expires: ");
									print_stamp(ss->expires);
									printf("\n");
									printf("        gate: {\n");
									for (g=0; g<XT_RTPPROXY_MAX_GATE; g++) {
										int st;
										printf("        %c: {\n", 'a'+g);
										for (st=0; st<2; st++) {
											printf("            %s: {port: %d, source: {ip: %s, port: %d}, learning:  \n", 
												st==0?"rtp":"rtcp",
												ss->gate[g].stream[st].port,
												ip2s(ss->gate[g].stream[st].source.ip),
												ss->gate[g].stream[st].source.port
											);
											print_stamp(ss->gate[g].stream[st].learning),
											printf(", last-packet: ");
											print_stamp(ss->gate[g].stream[st].last_packet_stamp);
											printf("}\n");
										}
										printf("            always-learn: %d\n", ss->gate[g].always_learn);
										printf("            stat: "); print_stat(&ss->gate[g].stat); printf("\n");
										
										printf("          }\n");
									}
									printf("        }\n");
									printf("      }\n");
								}
								printf("    }\n");
							}
							printf("  }\n");
						}
						printf("}\n");
				};
			}
			xt_RTPPROXY_release_switchboards(&handle, out_switchboard);
			break;
		}

		case cmdAlloc: {
			struct xt_rtpproxy_sockopt_switchboard out_switchboard_id;
			struct xt_rtpproxy_session out_session;
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_alloc_session(\n");
				print_in_switchboard(in_switchboard_fl?&in_switchboard:NULL); fprintf(stderr, ",\n");
				print_in_session(&in_session); fprintf(stderr, "\n");
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPROXY_alloc_session(&handle, in_switchboard_fl?&in_switchboard:NULL, &in_session, &out_switchboard_id, &out_session) < 0)
				goto err; 
			if (verbose) print_format(); 
			if (verbose) print_header(&handle);

			switch (output_format) {
				case fmtXml:
					printf("<switchboard>\n");
					printf("  <gate id=\"a\" ip=\"%s\" port=\"%d\" />\n", 
                				ip2s(out_switchboard_id.gate[0].addr.ip),
				                out_switchboard_id.gate[0].addr.port
				        ); /* ip2s ... result in static buffer */
					printf("  <gate id=\"b\" ip=\"%s\" port=\"%d\" />\n", 
                				ip2s(out_switchboard_id.gate[1].addr.ip),
				                out_switchboard_id.gate[1].addr.port
				        ); /* ip2s ... result in static buffer */
					printf("  <session sess_id=\"%d\">\n",
						out_session.sess_id
					);
					printf("    <created "); print_stamp(out_session.created); printf(" />\n");
					for (i=0; i<XT_RTPPROXY_MAX_GATE; i++) {
						printf("    <gate id=\"%c\">\n", 'a'+i);
						for (j=0; j<2; j++) {	
							printf("      <port proto=\"%s\" value=\"%d\" />\n", j?"rtcp":"rtp", out_session.gate[i].stream[j].port);

						}
						printf("    </gate>\n");
					}
					printf("  </session>\n");
					printf("</switchboard>\n");
					break;

				case fmtTab:
					printf("gate-a/ip\tgate-b/port\tgate-b/ip\tgate-b/port");
					printf("\tsession/sess-id\tsession/created/tick\tsession/created/datetime");
					printf("\tport-a/rtp\tport-a/rctp\tport-b/rtp\tport-b/rtcp\n");
					printf("%s\t%d", 
						ip2s(out_switchboard_id.gate[0].addr.ip),
                                                out_switchboard_id.gate[0].addr.port
					); /* ip2s ... result in static buffer */
					printf("\t%s\t%d", 
						ip2s(out_switchboard_id.gate[1].addr.ip), 
						out_switchboard_id.gate[1].addr.port
					); /* ip2s ... result in static buffer */
					printf("\t%d\t", out_session.sess_id);
					print_stamp(out_session.created);
					printf("\t%d\t%d\t%d\t%d\n",
						out_session.gate[0].stream[0].port,
						out_session.gate[0].stream[1].port,
						out_session.gate[1].stream[0].port,
						out_session.gate[1].stream[1].port
					);
					break;
				default:
					printf("switchboard: {gate-a: {ip: %s, port: %d}, ", 
						ip2s(out_switchboard_id.gate[0].addr.ip), 
						out_switchboard_id.gate[0].addr.port
					); /* ip2s ... result in static buffer */
					printf("gate-b: {ip: %s, port: %d}},\n", 
						ip2s(out_switchboard_id.gate[1].addr.ip), 
						out_switchboard_id.gate[1].addr.port
					); /* ip2s ... result in static buffer */
					printf("session: {sess-id: %d, created: ", 
						out_session.sess_id
					);
					print_stamp(out_session.created);
					printf(", port: {a: {rtp: %d, rtcp: %d}, b: {rtp: %d, rtcp: %d}}}\n",
						out_session.gate[0].stream[0].port,
						out_session.gate[0].stream[1].port,
						out_session.gate[1].stream[0].port,
						out_session.gate[1].stream[1].port
					);
					break;
			}

			break;
		}
		case cmdUpdate:
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_update_session(\n");
				print_in_switchboard(in_switchboard_fl?&in_switchboard:NULL); fprintf(stderr, ",\n");
				print_in_session_id(in_sess_id_fl?&in_sess_id:NULL); fprintf(stderr, ",\n");
				print_in_session(&in_session); fprintf(stderr, "\n");
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPROXY_update_session(&handle, in_switchboard_fl?&in_switchboard:NULL, in_sess_id_fl?&in_sess_id:NULL, &in_session) < 0)
				goto err;
			if (verbose) print_header(&handle);
			break;

		case cmdDelete:
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_delete_session(\n");
				print_in_switchboard(in_switchboard_fl?&in_switchboard:NULL); fprintf(stderr, ",\n");
				print_in_session_id(in_sess_id_fl?&in_sess_id:NULL); fprintf(stderr, ",\n");
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPROXY_delete_session(&handle, in_switchboard_fl?&in_switchboard:NULL, in_sess_id_fl?&in_sess_id:NULL) < 0)
				goto err; 
			if (verbose) print_header(&handle);

			break;
	}
	if (output_format == fmtXml) printf("</iptrtpproxy>\n");

destroy:
	if (verbose) fprintf(stderr, "xt_RTPPROXY_close()\n");
	xt_RTPPROXY_close(&handle);
	if (verbose) fprintf(stderr, "Done\n");
	return 0;

err:
	fprintf(stderr, "ERROR: %s (%d)\n", handle.err_str, handle.err_no);
	if (handle.sockfd >= 0) {
		if (verbose) fprintf(stderr, "xt_RTPPROXY_close()\n");
		xt_RTPPROXY_close(&handle);
	}
	if (verbose) fprintf(stderr, "Done\n");
	return 1;
}

#if !defined(NO_SHARED_LIBS) || NO_SHARED_LIBS==0
/* make compiler happy and give it missing symbols */
#include <iptables.h>

#if 0
void register_target(struct iptables_target *me) {
}
#else
void xtables_register_target(struct xtables_target *me) {
}
#endif

#if 0
void exit_error(enum exittype status, char *msg, ...)
#else
void exit_error(enum exittype status, const char *msg, ...)
#endif
{
	va_list args;
	
	va_start(args, msg);
	fprintf(stderr, "%s v%s: ", program_name, IPTRTPPROXY_VERSION);
	vfprintf(stderr, msg, args);
	va_end(args);
	fprintf(stderr, "\n");
	exit(status);
}
																		
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

int check_inverse(const char option[], int *invert, int *optind, int argc)
{
	if (option && strcmp(option, "!") == 0) {
		if (*invert)
			exit_error(PARAMETER_PROBLEM, "Multiple `!' flags not allowed");
		*invert = TRUE;
		if (optind) {
			*optind = *optind+1;
			if (argc && *optind > argc)
				exit_error(PARAMETER_PROBLEM, "no argument following `!'");
		}
		return TRUE;
	}
	return FALSE;
}

#endif

