/***************************************************************************
 *   (C) 2007 by Tomas Mandys <tomas.mandys at 2p dot cz>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation;             *
 *                                                                         *
 ***************************************************************************/

/*
 * This implements the RTPPROXY target, which enables you to setup
 * RTP proxy - handle RTP+RTCP streams through proxy to traverse NATs
 *
 */

#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/version.h>
#include <linux/netfilter.h>
#include <linux/netfilter/x_tables.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#include <net/netfilter/nf_conntrack_compat.h>
#else
#include <net/netfilter/nf_conntrack.h>
#endif
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter/xt_RTPPROXY.h>
#include <linux/netdevice.h>
#include <net/ip.h>
#include <asm/semaphore.h>
#include <linux/spinlock.h>
#include <linux/kernel.h>

#define PRINTK(level, format, args...) printk(level "xt_RTPPROXY: " format , ## args)

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,24)
#ifdef __KERNEL__
#define NF_IP_PRE_ROUTING NF_INET_PRE_ROUTING
#define NF_IP_LOCAL_IN NF_INET_LOCAL_IN
#define NF_IP_FORWARD NF_INET_FORWARD
#define NF_IP_LOCAL_OUT NF_INET_LOCAL_OUT
#define NF_IP_POST_ROUTING NF_INET_POST_ROUTING
#endif
#endif

#if 0
/* override DEBUGP to printk, just for development purposes */
/* Note: __FILE__, __FUNCTION__ macro may be used */
#define DEBUGP(format, args...) printk(KERN_DEBUG "xt_RTPPROXY: " format, ## args)
#else
#define DEBUGP(format, args...)
#endif

#if 0
/* override DEBUGPT to printk in do_target, just for development purposes */
/* Note: __FILE__, __FUNCTION__ macro may be used */
#define DEBUGPT(format, args...) printk(KERN_DEBUG "xt_RTPPROXY: " format, ## args)
#else
#define DEBUGPT(format, args...)
#endif



MODULE_LICENSE("GPL");
MODULE_AUTHOR("Tomas Mandys <tomas.mandys@2p.cz>");
MODULE_DESCRIPTION("x_tables RTPPROXY target module");
MODULE_ALIAS("ipt_RTPPROXY");

/* return number of opposite gate, i.e. OPPOSITE_GATE(0 i.e. A) is 1 i.e B etc. */
#define OPPOSITE_GATE(i) (i^1)

/* serializes interrupt/softirq/user-context manipulation
 * - switchboard chain manipulation (switchboard_reg, prev, next)
 * - xt_rtpproxy_session.gate[].stream[]
 * - xt_rtpproxy_session.gate[].stat
 * - xt_rtpproxy_session.state
 * - xt_rtpproxy_session.expires
 * - xt_rtpproxy_sockopt_switchboard.gate[].stat
 * - xt_rtpproxy_sockopt_switchboard.stat.state
 */
static DEFINE_SPINLOCK(xt_lock);

/* serializes user-context access
 * - switchboard allocations and deallocations + chain manipulation
 * - xt_rtpproxy_sockopt_switchboard.stat.counter
 * - xt_rtpproxy_sockopt_out_config
 * - xt_rtpproxy_session.flags
 * - xt_rtpproxy_sockopt_switchboard.flags
 */
static DECLARE_MUTEX(xt_usercontext_mutex);

static struct xt_rtpproxy_switchboard *switchboard_reg = NULL;

static struct xt_rtpproxy_sockopt_out_config so_global;


/* interrupt-context, softirq?? */
static unsigned int do_target(
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
				struct sk_buff **pskb,
#else
				struct sk_buff *skb,
#endif
				const struct net_device *in,
				const struct net_device *out,
				unsigned int hooknum,
				const struct xt_target *target,
				const void *targinfo
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
				, void *userinfo
#endif
					 )
{
	const struct xt_rtpproxy_target_info *info = targinfo;
	struct iphdr *piph;
	struct udphdr *pudph;
#if defined(CONFIG_IP_NF_CONNTRACK) || defined(CONFIG_IP_NF_CONNTRACK_MODULE)
	struct ip_conntrack *ct;
	u_int32_t ctinfo;
	#define ASSERT IP_NF_ASSERT
#else
	struct nf_conn *ct;
	enum ip_conntrack_info ctinfo;
	#define ASSERT NF_CT_ASSERT
#endif
	u_int16_t port, *old_port, new_port;
	u_int32_t ip, *old_ip, new_ip;
	int input_gate, sess_no, is_rtcp;
	struct xt_rtpproxy_session *sess;
	time_t tstamp_off_sec;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)	
	struct sk_buff *skb = *pskb;
#endif

	/* may be called only for UDP packet, yes I'm sure, see target registration later */

	DEBUGPT("do_target(skb=%p, in=%p, out=%p, hooknum=%u, targetinfo=%p)\n", skb, in, out, hooknum, targinfo);
	/* Now we take destination ip:port from tuple, we'll do it instead of skb->nh.iph because
	 * it's non manipulable part of tuple and it has correct value in both PRE_ROUTING/LOCAL_OUT
	 * and POST_ROUTING (where &skb->nh.iph->daddr is already rewritten and is useless
	 * for identification of RTP session). It cannot be a reply (stream in opposite direction)
	 * beacuse the replies have different tuple */
	ASSERT(hooknum==NF_IP_PRE_ROUTING || hooknum==NF_IP_LOCAL_OUT || hooknum==NF_IP_POST_ROUTING || hooknum==NF_IP_LOCAL_IN);
#if defined(CONFIG_IP_NF_CONNTRACK) || defined(CONFIG_IP_NF_CONNTRACK_MODULE)
	DEBUGPT("do_target: ip_conntract_get()\n");
	ct = ip_conntrack_get(skb, &ctinfo);
	#define GET_TUPLE_IP(_dir_,_dest_) ct->tuplehash[_dir_].tuple._dest_.ip
#else
	DEBUGPT("do_target: nf_ct_get()\n");
	ct = nf_ct_get(skb, &ctinfo);
	#define GET_TUPLE_IP(_dir_,_dest_) ct->tuplehash[_dir_].tuple._dest_.u3.ip
#endif
	if (!ct) {
		DEBUGPT("do_target: ct is null (bad checksum?), dropping\n");
		return NF_DROP;
	}
	/* Connection must be valid and non reply - replies go in via opposite gate as originals, but allow CONTINUE not to block common communication not related to RTP, e.g. DNS */
	if (ctinfo >= IP_CT_IS_REPLY) {
		DEBUGPT("do_target: ctinfo >= IP_CT_IS_REPLY, IPF_CONTINUE\n");
		return XT_CONTINUE;
	}
	
	/* get location to rewrite IP:port of UDP header, actually we need only first 8 bytes */
	piph = (void *)skb->data;
	pudph = (void *)(skb->data + piph->ihl*4);
	DEBUGPT("do_target: GET_TUPLE_IP\n");
	ip = GET_TUPLE_IP(IP_CT_DIR_ORIGINAL, dst);
	port = ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u.udp.port);

	if (hooknum == NF_IP_PRE_ROUTING) {
		/* when all stuff is running on one box 2xRTP client+RTPPROXY (devel purposes only ?)
		 * then after POSTROUTING follows PREROUTING immediately and netfilter
		 * leaves skb->nfct untouched and RTPPROXY is mystified, source IP:port is address of opposite gate.
		 * conntrack should be invalidated and refreshed
		 */
		if (ip != piph->daddr ||
		    port != ntohs(pudph->dest)) {
			DEBUGPT("do_target: non-corresponding conntrack and dest address, local RTP routing?\n");
			return NF_DROP;

		#if 0
			This dummy devel code is left because I expect the netfilter could behave
			correctly in future also for locally rerouted packets or if somebody
			would like to improve code not to repeat my unsucessfull attempts with kernel 2.6.17.8

			1) nf_conntrack_put causes kernel crash

			/* force conntrack refresh */
			DEBUGPT("do_target: refreshing conntrack\n");
			nf_conntrack_put(skb->nfct);
			nf_ct_untrack(skb);
			return XT_CONTINUE;		

			2) huge hack and ip_conntrack_in is not external symbol

			/* call early conntract callback */
			skb->nfct = NULL;
			it's nt possible to call ip_conntrack_in from module and I don't other way how to refresh conntrack
			if (ip_conntrack_in(hooknum, pskb, in, out, NULL) == NF_DROP) {
				return NF_DROP;
			}
			ct = ip_conntrack_get(skb, &ctinfo);
			DEBUGPT("do_target: ctinfo >= IP_CT_IS_REPLY, IPF_CONTINUE\n");
			if (ctinfo >= IP_CT_IS_REPLY) {
				return XT_CONTINUE;
			}
			/* get location to rewrite IP:port of UDP header, actually we need only first 8 bytes */
			pudph = (void *)(skb->data + piph->ihl*4);
		#endif

		}
	}	

	ASSERT(ct && (ctinfo < IP_CT_IS_REPLY));
	DEBUGPT("do_target:{ct->src: %d.%d.%d.%d:%u, ct->dst: %d.%d.%d.%d:%u} reply:{ct->src: %d.%d.%d.%d:%u, ct->dst: %d.%d.%d.%d:%u}, ctinfo: %u\n",
		NIPQUAD(GET_TUPLE_IP(IP_CT_DIR_ORIGINAL, src)), ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.u.udp.port),
		NIPQUAD(GET_TUPLE_IP(IP_CT_DIR_ORIGINAL, dst)), ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u.udp.port),
		NIPQUAD(GET_TUPLE_IP(IP_CT_DIR_REPLY, src)), ntohs(ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.u.udp.port),
		NIPQUAD(GET_TUPLE_IP(IP_CT_DIR_REPLY, dst)), ntohs(ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.u.udp.port),
		ctinfo);

	DEBUGPT("do_target:skb->iph.src: %d.%d.%d.%d:%u, skb->iph.dst: %d.%d.%d.%d:%u\n",
		NIPQUAD(piph->saddr), ntohs(pudph->source),
		NIPQUAD(piph->daddr), ntohs(pudph->dest)
		);
	spin_lock_bh(&xt_lock);
	/* ip:port must be a port related to gate_a/b */
	for (input_gate=0, sess=NULL; input_gate < XT_RTPPROXY_MAX_GATE; input_gate++) {
		if (ip == info->gate[input_gate].ip) {
			if (port >= info->gate[input_gate].port && port < info->gate[input_gate].port+2*info->max_sess) {
				sess_no = (port-info->gate[input_gate].port) >> 1;
				is_rtcp = (port-info->gate[input_gate].port) & 0x01; /* TODO: check if may be also even port number */
				sess = info->switchboard->sessions + sess_no;
				break;
			}
		}
	}
	if (!sess) {
		spin_unlock_bh(&xt_lock);
		DEBUGPT("do_target: session not found, XT_CONTINUE\n");
		return XT_CONTINUE;
	}

	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
	tstamp_off_sec = skb->tstamp.off_sec;
	#else
	{
		struct timeval stamp;
		skb_get_timestamp(skb, &stamp);
		tstamp_off_sec = stamp.tv_sec;
	}
	#endif

	DEBUGPT("do_target: session found\n");
	if (!tstamp_off_sec)
		__net_timestamp(skb);

	switch (hooknum) {
		case NF_IP_PRE_ROUTING:
		case NF_IP_LOCAL_OUT:
			/* check if session is active */
			switch (sess->state) {
				case xt_rtpproxy_INIT1:
				case xt_rtpproxy_INIT2:
				case xt_rtpproxy_FORWARD1:
				case xt_rtpproxy_FORWARD2:
					break;
				default:
					spin_unlock_bh(&xt_lock);
					DEBUGPT("do_target: sess->state=%d, NF_DROP\n", sess->state);
					return NF_DROP;
			}

			/* check if session is still active */
			if (sess->expires < tstamp_off_sec) {
				info->switchboard->so.stat.state[sess->state]--;
				sess->state = xt_rtpproxy_EXPIRED;
				info->switchboard->so.stat.state[sess->state]++;
				spin_unlock_bh(&xt_lock);
				DEBUGPT("do_target: session expired, NF_DROP\n");
				return NF_DROP;
			}

			if (sess->gate[input_gate].stream[is_rtcp].source.ip != piph->saddr || sess->gate[input_gate].stream[is_rtcp].source.port != ntohs(pudph->source)) {
				if (sess->gate[input_gate].stream[is_rtcp].learning >= tstamp_off_sec) {
					/* learn source address:port */
					sess->gate[input_gate].stream[is_rtcp].source.ip = piph->saddr;
					sess->gate[input_gate].stream[is_rtcp].source.port = ntohs(pudph->source);
					if (!sess->gate[input_gate].always_learn)
						sess->gate[input_gate].stream[is_rtcp].learning = 0;  /* learn only once */
					/* sess->gate[input_gate].learnt = skb->tstamp; */
				}
				else {
					/* drop unlearnable packet silently */
					spin_unlock_bh(&xt_lock);
					DEBUGPT("do_target: unlearnable packet, NF_DROP\n");
					return NF_DROP;
				}
			}
			/* rewrite destination address to opposite side (learnt/forced IP:port) */
			if (sess->gate[OPPOSITE_GATE(input_gate)].stream[is_rtcp].source.ip == 0) {
				/* not learnt/received any packet from opposite side, we cannot forward it */
				if (sess->gate[OPPOSITE_GATE(input_gate)].stream[is_rtcp].learning >= tstamp_off_sec) {
					/* if we don't missed chance forever to learn opposite address then prolong expires even we cannot forward the packet */
					sess->expires = tstamp_off_sec+info->expires_timeout;
				}
				spin_unlock_bh(&xt_lock);
				DEBUGPT("do_target: no packet received from opposite side, NF_DROP\n");
				return NF_DROP;
			}
			switch (sess->state) {
				case xt_rtpproxy_INIT2:
					info->switchboard->so.stat.state[sess->state]--;
					sess->state = xt_rtpproxy_FORWARD1; /* first forwarded packet */
					info->switchboard->so.stat.state[sess->state]++;
					break;
				case xt_rtpproxy_FORWARD1:
					if (sess->gate[OPPOSITE_GATE(input_gate)].stream[is_rtcp].last_packet_stamp) {
						/* we've already got packet from opposite side (of the same stream!) */
						info->switchboard->so.stat.state[sess->state]--;
						sess->state = xt_rtpproxy_FORWARD2;
						info->switchboard->so.stat.state[sess->state]++;
					}
					break;
			}

			new_ip = sess->gate[OPPOSITE_GATE(input_gate)].stream[is_rtcp].source.ip;
			new_port = sess->gate[OPPOSITE_GATE(input_gate)].stream[is_rtcp].source.port;

			sess->gate[input_gate].stat.packets_in++;
			sess->gate[input_gate].stat.bytes_in += pudph->len;
			info->switchboard->so.gate[input_gate].stat.packets_in++;
			info->switchboard->so.gate[input_gate].stat.bytes_in += pudph->len;
			sess->expires = tstamp_off_sec+info->expires_timeout;
			sess->gate[input_gate].stream[is_rtcp].last_packet_stamp = tstamp_off_sec;
			break;

		case NF_IP_POST_ROUTING:
		case NF_IP_LOCAL_IN:
			/* rewrite source address to proxy gate */
			new_ip = info->gate[OPPOSITE_GATE(input_gate)].ip;
			new_port = info->gate[OPPOSITE_GATE(input_gate)].port+2*sess_no+is_rtcp;

			sess->gate[OPPOSITE_GATE(input_gate)].stat.packets_out++;
			sess->gate[OPPOSITE_GATE(input_gate)].stat.bytes_out += pudph->len;
			info->switchboard->so.gate[OPPOSITE_GATE(input_gate)].stat.packets_out++;
			info->switchboard->so.gate[OPPOSITE_GATE(input_gate)].stat.bytes_out += pudph->len;
			break;

		default: /* never happens */
			spin_unlock_bh(&xt_lock);
			DEBUGPT("do_target: unexpected hooknum = %d, XT_CONTINUE\n", hooknum);
			return XT_CONTINUE;
	}
	spin_unlock_bh(&xt_lock);

	if (!skb_make_writable(
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
		pskb, 
	#else
		skb,
	#endif
		piph->ihl*4 + sizeof(*pudph))) { /* it changes skb !!! */
		DEBUGPT("do_target: !skb_make_writable\n");
                return NF_DROP;
	}

	piph = (void *)skb->data;
	pudph = (void *)(skb->data + piph->ihl*4);
	switch (hooknum) {
		case NF_IP_PRE_ROUTING:
		case NF_IP_LOCAL_OUT:
			old_ip = &piph->daddr;
			old_port = &pudph->dest;
			/* dst_release(skb->dst); idea taken from NAT but does not work
			skb->dst = NULL; */
			break;
		case NF_IP_POST_ROUTING:
		case NF_IP_LOCAL_IN:
			old_ip = &piph->saddr;
			old_port = &pudph->source;
			break;
		default:
			return NF_DROP;  /* never happen */
	}
	#if 0
	if (hooknum == NF_IP_LOCAL_OUT) {
		/* bug: seems checksum is incorrect for locally generated udp packets, e.g. data not counted, seems also ^0xFFFFed (kernel 2.6.17.8)  */
		pudph->check = 0;
	}
	#endif

	/* rewrite UDP checksum if provided */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
	if (pudph->check) {
		pudph->check = ip_nat_cheat_check(~*old_ip, new_ip, 
				ip_nat_cheat_check(*old_port ^ 0xFFFF, new_port,
				pudph->check));
	}
	/* rewrite also IP header checksum */
	piph->check = ip_nat_cheat_check(~*old_ip, new_ip, piph->check);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
        if (pudph->check || skb->ip_summed == CHECKSUM_PARTIAL) {
	
		pudph->check = nf_proto_csum_update(skb, ~*old_ip, new_ip, pudph->check, 1);
		pudph->check = nf_proto_csum_update(skb, *old_port ^ 0xFFFF, new_port, pudph->check, 0);
	}
	/* rewrite also IP header checksum */
	piph->check = nf_csum_update(~*old_ip, new_ip, piph->check);
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,6,24)
	if (pudph->check || skb->ip_summed == CHECKSUM_PARTIAL) {
		inet_proto_csum_replace4(&pudph->check, skb, *old_ip, new_ip, 1);
		inet_proto_csum_replace2(&pudph->check, skb, *old_port, new_port, 0);
		if (!pudph->check)
			pudph->check = CSUM_MANGLED_0;
	}
	/* rewrite also IP header checksum */
	csum_replace4(&piph->check, *old_ip, new_ip);
#else
	if (pudph->check || skb->ip_summed == CHECKSUM_PARTIAL) {
		nf_proto_csum_replace4(&pudph->check, skb, *old_ip, new_ip, 1);
		nf_proto_csum_replace2(&pudph->check, skb, *old_port, new_port, 0);
		if (!pudph->check)
			pudph->check = CSUM_MANGLED_0;
	}
	/* rewrite also IP header checksum */
	nf_csum_replace4(&piph->check, *old_ip, new_ip);
#endif

	DEBUGPT("do_target: old_addr: %d.%d.%d.%d:%u, new_addr: %d.%d.%d.%d:%u, udph->check: %u, ip_summed: %d\n", NIPQUAD(*old_ip), ntohs(*old_port), NIPQUAD(new_ip), new_port, pudph->check, skb->ip_summed);
	pudph->check = 0; /* avoid UDP checksum, I'm not able to calculate it correctly */

	*old_port = htons(new_port);
	*old_ip = new_ip;

	/* skb->nfcache |= NFC_ALTERED; */
	DEBUGPT("do_target: NF_ACCEPT\n");
	return NF_ACCEPT;
}

/* recheck session status and calc statistics by state, xt_lock+xt_usercontext_mutex must be aquired */
static inline void xt_do_session_audit(struct timeval *stamp, struct xt_rtpproxy_switchboard *switchboard, struct xt_rtpproxy_session *sess)
{
	switch (sess->state) {
		case xt_rtpproxy_INIT1:
		case xt_rtpproxy_INIT2:
		case xt_rtpproxy_FORWARD1:
		case xt_rtpproxy_FORWARD2:
			if (sess->expires < stamp->tv_sec) {
				switchboard->so.stat.state[sess->state]--;
				sess->state = xt_rtpproxy_EXPIRED;
				switchboard->so.stat.state[sess->state]++;
				switchboard->so.stat.counter.expired++;
			}
			break;
		default:
			break;
	}
	switch (sess->state) {
		case xt_rtpproxy_EXPIRED:
		case xt_rtpproxy_DESTROYED:
			if (sess->expires + switchboard->so.resurrection_timeout < stamp->tv_sec) {
				switchboard->so.stat.state[sess->state]--;
				sess->state = xt_rtpproxy_NONE;
				switchboard->so.stat.state[sess->state]++;
			}
		default: ;
	}
}

/* userspace related communication, xt_usercontext_mutex must be aquired */
static inline struct xt_rtpproxy_switchboard* xt_find_switchboard(struct xt_rtpproxy_sockopt_in_switchboard *switchboard, int *opposite_flag)
{
	struct xt_rtpproxy_switchboard *p;
	int j, port_fl;

	/* this is helper, if ip-a==ip-b then port-a must not be equal to port.b but because client may not know port-b
	   we'll find it */
	port_fl = switchboard->gate[0].ip == switchboard->gate[OPPOSITE_GATE(0)].ip &&
		  switchboard->gate[0].port == switchboard->gate[OPPOSITE_GATE(0)].port;

	for (p = switchboard_reg; p ; p = p->next ) {
		for (j=0; j<XT_RTPPROXY_MAX_GATE; j++) {
			if ( p->so.gate[0].addr.ip == switchboard->gate[j].ip &&
			     p->so.gate[0].addr.port == switchboard->gate[j].port &&
			     p->so.gate[OPPOSITE_GATE(0)].addr.ip == switchboard->gate[OPPOSITE_GATE(j)].ip &&
			     (p->so.gate[OPPOSITE_GATE(0)].addr.port == switchboard->gate[OPPOSITE_GATE(j)].port || port_fl) ) {
				switchboard->gate[OPPOSITE_GATE(j)].port = p->so.gate[OPPOSITE_GATE(0)].addr.port;  /* meaningful when port_fl */
				*opposite_flag = (j != 0);
				return p;
			}
		}
	}
	return NULL;
}

/* there is no uniq cmd numbering space, 2 separates for GET and SET */
#define XT_OPT_SET_CMD 0x8000

/* user context code */
static int xt_do_get_ctl(struct sock *sk, int cmd, void *user, int *len)
{
	struct xt_rtpproxy_sockopt_header in_header;
	struct xt_rtpproxy_sockopt_in_switchboard in_switchboard;
	struct xt_rtpproxy_sockopt_in_sess_id in_sess_id;
	struct timeval stamp;
	int ret = -EFAULT;
	int in_len;
	void *user2;

	DEBUGP("xt_do_%cet_ctl(sk=%p, cmd=%d, user=%p, len=%u)\n", (cmd&XT_OPT_SET_CMD)?'s':'g', sk, cmd & ~XT_OPT_SET_CMD, user, *len);
	/* acquire lock here/early because we ensapsulate also statistics changes */
	/* we cannot use spin_lock because copy_from/to_user calls sleep! */
	if (down_interruptible(&xt_usercontext_mutex)) {
		ret = -EINTR;
		goto ret_no_unlock;
	}

	if (!capable(CAP_NET_ADMIN)) {
		so_global.stat.perm++;
		ret = -EPERM;
		goto ret_no_unlock;
	}

	user2 = user;
	in_len = *len;
	if (in_len == 0) {
		memset(&in_header, 0, sizeof(in_header));
	}
	else {
		if (in_len < sizeof(in_header))
			goto fault;
		if (copy_from_user (&in_header, user2, sizeof(in_header))) {
			goto fault;
		}
		user2 += sizeof(in_header);
		in_len -= sizeof(in_header);
	}
	DEBUGP("xt_do_get_ctl: in_header={.version=0x%x, .flags=0x%x}\n", in_header.version, in_header.flags);
	if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD) {
		if (in_len < sizeof(in_switchboard))
			goto fault;
		if (copy_from_user (&in_switchboard, user2, sizeof(in_switchboard))) {
			goto fault;
		}
		user2 += sizeof(in_switchboard);
		in_len -= sizeof(in_switchboard);
		DEBUGP("xt_do_get_ctl: in_switchboard={.gate-a=0x%x:%u, .gate-b=0x%x:%u}\n", in_switchboard.gate[0].ip, in_switchboard.gate[0].port, in_switchboard.gate[OPPOSITE_GATE(0)].ip, in_switchboard.gate[OPPOSITE_GATE(0)].port);
	}
	if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_SESS_ID) {
		if (in_len < sizeof(in_sess_id))
			goto fault;
		if (copy_from_user (&in_sess_id, user2, sizeof(in_sess_id))) {
			goto fault;
		}
		user2 += sizeof(in_sess_id);
		in_len -= sizeof(in_sess_id);
		DEBUGP("xt_do_get_ctl: in_sess_id={.min=%u, .max=%u, .created=%u}\n", in_sess_id.sess_id_min, in_sess_id.sess_id_max, in_sess_id.created);
	}
	else {
		in_sess_id.sess_id_min = 0;
		in_sess_id.sess_id_max = 0xFFFF;
		in_sess_id.created = 0;
	}

	if (in_header.version > XT_RTPPROXY_SOCKOPT_VERSION) {
		PRINTK(KERN_WARNING, "xt_get_ctl: bad version %d>%d\n", in_header.version, XT_RTPPROXY_SOCKOPT_VERSION);
		goto fault;
	}
	do_gettimeofday(&stamp);
	
	switch (cmd) {
		case XT_RTPPROXY_SO_GET_CONFIG: {
			struct xt_rtpproxy_switchboard *p;
			int opposite_flag;
			struct xt_rtpproxy_sockopt_num_session out_count;

			if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD) {
				p = xt_find_switchboard(&in_switchboard, &opposite_flag);
			}
			else {
				p = switchboard_reg;
			}

			if (p && (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_COUNT)) {
				if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD) {
					out_count.num_switchboard = 1;
					out_count.num_session = p->so.max_sess;
				}
				else
					out_count = so_global.total;
			}
			else {
				memset(&out_count, 0, sizeof(out_count));
			}
			if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_COUNT)
				so_global.h.flags |= XT_RTPPROXY_SOCKOPT_FLAG_OUT_COUNT;
			else
				so_global.h.flags &= ~XT_RTPPROXY_SOCKOPT_FLAG_OUT_COUNT;

			if (p && in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD)
				so_global.h.flags |= XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD;
			else
				so_global.h.flags &= ~XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD;
			
			user2 = user;
			DEBUGP("xt_do_get_ctl: copy_to_user(so_global={.flags=0x%x})\n", so_global.h.flags);
			if (copy_to_user(user2, &so_global, sizeof(so_global))) {
				goto brk;
			}
			user2 += sizeof(so_global);
			if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_COUNT) {
				DEBUGP("xt_do_get_ctl: copy_to_user(out_count)\n");
				if (copy_to_user(user2, &out_count, sizeof(out_count))) {
					goto brk;
				}
				user2 += sizeof(out_count);
			}
			for (; p; p=p->next) {
				int sess_id, sess_id_max, sess_flag;

				spin_lock_bh(&xt_lock);
				sess_id_max = ((int)in_sess_id.sess_id_max+1 >= p->so.max_sess)? p->so.max_sess : ((int)in_sess_id.sess_id_max+1);
				if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_FORCE_SWITCHBOARD_AUDIT) {
					int i;
					for (i=in_sess_id.sess_id_min; i<sess_id_max; i++) {
						if (!in_sess_id.created || p->sessions[i].created == in_sess_id.created)
							xt_do_session_audit(&stamp, p, p->sessions+i);
					}
				}

				/* find first session having valuable data, i.e. skip not yet used sessions */
				sess_flag = 0;
				if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION) {
					for (sess_id = in_sess_id.sess_id_min; sess_id < sess_id_max; sess_id++) {
						if (in_sess_id.created && p->sessions[sess_id].created != in_sess_id.created)
							continue; /* created does not match */
						if (p->sessions[sess_id].state != xt_rtpproxy_NONE) {
							sess_flag = 1;
							break;
						}
					}
				}

				if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD) {
					/* save kernel stack space, variable is on the stack in the block only, I hope */
					struct xt_rtpproxy_sockopt_switchboard so_copy;

					/* we must copy p->so because there are some fields requiring xt_lock and copy_to_user calls sleep */
					memcpy(&so_copy, &p->so, sizeof(so_copy));
					spin_unlock_bh(&xt_lock);

					if (!p->next || in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD)
						so_copy.flags &= ~XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD;
					else
						so_copy.flags |= XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD;

					if (sess_flag)
						so_copy.flags |= XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION;
					else
						so_copy.flags &= ~XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION;
					DEBUGP("xt_do_get_ctl: flags=0x%x, copy_to_user(so_copy)\n", so_copy.flags);
					if (copy_to_user(user2, &so_copy, sizeof(so_copy))) {
						goto brk;
					}
					user2 += sizeof(so_copy);
				}
				else {
					spin_unlock_bh(&xt_lock);
				}

				if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_RESET_SWITCHBOARD_STAT) {
					memset(&p->so.stat.counter, 0, sizeof(p->so.stat.counter));
				}
				if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_RESET_SWITCHBOARD_PACKET_STAT) {
					int i;
					spin_lock_bh(&xt_lock);
					for (i=0; i<XT_RTPPROXY_MAX_GATE; i++) {
						memset(&p->so.gate[i].stat, 0, sizeof(p->so.gate[i].stat));
					}
					spin_unlock_bh(&xt_lock);
				}
				
				for (sess_id = in_sess_id.sess_id_min; sess_id < sess_id_max; sess_id++) {

					if (in_sess_id.created && p->sessions[sess_id].created != in_sess_id.created)
						continue; /* created does not match */
					if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION && sess_flag) {
						/* a session might change state from xt_rtpproxy_NONE but previous
						   flags says "no more session"
						   but cannot change to xt_rtpproxy_NONE
						 */
						int j;
						struct xt_rtpproxy_session sess_copy;

						spin_lock_bh(&xt_lock);
						/* we must copy p->so because there are some fields requiring xt_lock and copy_to_user calls sleep */
						if (p->sessions[sess_id].state != xt_rtpproxy_NONE) {

							memcpy(&sess_copy, p->sessions+sess_id, sizeof(sess_copy));
							sess_copy.next = sess_copy.prev = 0; /* invalidate pointers */
							/* find next session, i.e. skip not important ones */
							sess_flag = 0;
							for (j = sess_id+1; j < sess_id_max; j++) {
								if (p->sessions[j].state != xt_rtpproxy_NONE) {
									sess_flag = 1;
									break;
								}
							}
							spin_unlock_bh(&xt_lock);
							if (sess_flag)
								sess_copy.flags |= XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION;
							else
								sess_copy.flags &= ~XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION;

							DEBUGP("xt_do_get_ctl: sess_id=%d, flags=0x%x, copy_to_user(sess_copy)\n", sess_id, sess_copy.flags);
							if (copy_to_user(user2, &sess_copy, sizeof(sess_copy))) {
								goto brk;
							}
							user2 += sizeof(sess_copy);
						}	
						else {
							spin_unlock_bh(&xt_lock);
						}
					}
					if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_RESET_SWITCHBOARD_PACKET_STAT) {
						int i;
						spin_lock_bh(&xt_lock);
						for (i=0; i<XT_RTPPROXY_MAX_GATE; i++) {
							memset(&p->sessions[sess_id].gate[i].stat, 0, sizeof(p->sessions[sess_id].gate[i].stat));
						}
						spin_unlock_bh(&xt_lock);
					}

				}
				if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD)
					break;  /* only one gate required */
			}
			ret = 0;
			break;
		}
		case XT_RTPPROXY_SO_GET_ALLOC_SESSION: {
			struct xt_rtpproxy_switchboard *p;
			struct xt_rtpproxy_sockopt_in_alloc_session in_sess;
			struct xt_rtpproxy_session *sess = NULL;
			struct xt_rtpproxy_sockopt_header out_header;
			struct xt_rtpproxy_sockopt_switchboard so_copy;
			struct xt_rtpproxy_session sess_copy;

			int i, j, opposite_flag = 0;

			if (in_len < sizeof(in_sess))
				goto brk;
			if (copy_from_user (&in_sess, user2, sizeof(in_sess))) {
				goto brk;
			}
			user2 += sizeof(in_sess);
			in_len -= sizeof(in_sess);
			DEBUGP("xt_do_get_ctl: in_sess={"
				".source[A]={"
					"RTP={.addr=0x%x:%u, .learning_timeout=%u, .flags=%x},"
					"RTCP={.addr=0x%x:%u, .learning_timeout=%u, .flags=%x},"
					".always_learn=%u},"
				".source[B]={"
					"RTP={.addr=0x%x:%u, .learning_timeout=%u, .flags=%x},"
					"RTCP={.addr=0x%x:%u, .learning_timeout=%u, .flags=%x},"
					".always_learn=%u}"
				"}\n",
				in_sess.source[0].stream[0].addr.ip, in_sess.source[0].stream[0].addr.port, in_sess.source[0].stream[0].learning_timeout, in_sess.source[0].stream[0].flags,
				in_sess.source[0].stream[1].addr.ip, in_sess.source[0].stream[1].addr.port, in_sess.source[0].stream[1].learning_timeout, in_sess.source[0].stream[1].flags,
				in_sess.source[0].always_learn,
				in_sess.source[1].stream[0].addr.ip, in_sess.source[1].stream[0].addr.port, in_sess.source[1].stream[0].learning_timeout, in_sess.source[1].stream[0].flags,
				in_sess.source[1].stream[1].addr.ip, in_sess.source[1].stream[1].addr.port, in_sess.source[1].stream[1].learning_timeout, in_sess.source[1].stream[1].flags,
				in_sess.source[1].always_learn
			);

			p = (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD)?xt_find_switchboard(&in_switchboard, &opposite_flag):switchboard_reg;
			spin_lock_bh(&xt_lock);
			while (p) {
				int n;
				/* we go around whole session range to decrease chance of cross talks */
				for (n=p->so.max_sess; n > 0; n--) {
					sess = p->sessions + p->next_sess_id;
					xt_do_session_audit(&stamp, p, sess);
					if (p->next_sess_id < p->so.max_sess-1)
						p->next_sess_id++;
					else
						p->next_sess_id = 0;
					if (sess->state == xt_rtpproxy_NONE)
						goto alloc_found_sess;
				}

				if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD || !p->next) {
					p->so.stat.counter.alloc_failed++;
					break;
				}
				p = p->next;
			}
			spin_unlock_bh(&xt_lock);
			DEBUGP("xt_do_get_ctl: cannot alloc session\n");
			goto brk;

		alloc_found_sess:
			/* set session data and adjust status and prepare out data*/
			p->so.stat.state[sess->state]--;
			sess->state = xt_rtpproxy_NONE;
			sess->created = stamp.tv_sec;
			sess->expires = stamp.tv_sec + p->so.expires_timeout;
			DEBUGP("xt_do_get_ctl: allocating sess_id=%d, created=%u, expires=%u\n", sess->sess_id, sess->created, sess->expires);
			for (i=0; i<XT_RTPPROXY_MAX_GATE; i++) {
				int k;
				/* if forced switchboard that was demanded in opposite order we must assign in_sess[0]->sess[1], in_sess[1]->sess[0] */
				k = opposite_flag?XT_RTPPROXY_MAX_GATE-i-1:i;
				for (j=0; j<2; j++) {
					if (in_sess.source[k].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR)
						sess->gate[i].stream[j].source = in_sess.source[k].stream[j].addr;
					else
						memset(&sess->gate[i].stream[j].source, 0, sizeof(sess->gate[i].stream[j].source));

					if (in_sess.source[k].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT &&
					    in_sess.source[k].stream[j].learning_timeout) {
						sess->gate[i].stream[j].learning = stamp.tv_sec+in_sess.source[k].stream[j].learning_timeout;
						if (sess->gate[i].stream[j].learning > sess->expires)
							sess->expires = sess->gate[i].stream[j].learning;
					}
					else
						sess->gate[i].stream[j].learning = 0;
					if (sess->gate[i].stream[j].learning || sess->gate[i].stream[j].source.ip)
						sess->state = xt_rtpproxy_INIT1;
				}	
				sess->gate[i].always_learn = in_sess.source[k].always_learn;
				memset(&sess->gate[i].stat, 0, sizeof(sess->gate[i].stat));			
			}
			/* maybe both sides initialized ? */
			for (j=0; j<2; j++) {
				for (i=0; i<XT_RTPPROXY_MAX_GATE; i++) {
					if (!sess->gate[i].stream[j].learning && !sess->gate[i].stream[j].source.ip)
						break;
				}
				if (i >= XT_RTPPROXY_MAX_GATE) { /* all source IPs are known/learning for at least one stream ? */
					sess->state = xt_rtpproxy_INIT2;
					break;
				}
			}
			p->so.stat.state[sess->state]++;
			if (sess->state == xt_rtpproxy_NONE) { /* we need at least on source addr or learning active */
				spin_unlock_bh(&xt_lock);
				goto brk;
			}
			memcpy(&so_copy, &p->so, sizeof(so_copy));
			memcpy(&sess_copy, sess, sizeof(sess_copy));
			sess_copy.next = sess_copy.prev = 0; /* invalidate pointers */
			spin_unlock_bh(&xt_lock);


			p->so.stat.counter.alloc++; /* it's allocated, does not matter if theoretically copy_to_user() fails */

			out_header.version = XT_RTPPROXY_SOCKOPT_VERSION;
			out_header.flags = XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD;
			DEBUGP("xt_do_get_ctl: copy_to_user(out_header={.flags=0x%x})\n", out_header.flags);
			user2 = user;
			if (copy_to_user(user2, &out_header, sizeof(out_header))) {
				goto brk;
			}
			user2 += sizeof(out_header);
			so_copy.flags |= XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION;
			so_copy.flags &= ~XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD;
			DEBUGP("xt_do_get_ctl: flags=0x%x, copy_to_user(so_copy})\n", so_copy.flags);
			if (copy_to_user(user2, &so_copy, sizeof(so_copy))) {
				goto brk;
			}
			user2 += sizeof(so_copy);
			sess_copy.flags &= ~XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION;
			DEBUGP("xt_do_get_ctl: flags=0x%x, copy_to_user(sess_copy})\n", sess_copy.flags);
			if (copy_to_user(user2, &sess_copy, sizeof(sess_copy))) {
				goto brk;
			}
			user2 += sizeof(sess_copy);
			ret = 0;
			break;
		}
		case XT_RTPPROXY_SO_SET_UPDATE_SESSION|XT_OPT_SET_CMD: {
			struct xt_rtpproxy_switchboard *p;
			struct xt_rtpproxy_sockopt_in_alloc_session in_sess;
			struct xt_rtpproxy_session *sess = NULL;
			int sess_id;
			int i, j, opposite_flag = 0;

			if ((in_header.flags & (XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD)) == 0)
				goto brk; /* we need both switchboard and sess_id */
			p = xt_find_switchboard(&in_switchboard, &opposite_flag);
			if (!p) {
				DEBUGP("xt_do_get_ctl: switchboard not found\n");
				goto brk;
			}
			DEBUGP("xt_do_get_ctl: switchboard found\n");
			if (in_len < sizeof(in_sess))
				goto fault;
			if (copy_from_user (&in_sess, user2, sizeof(in_sess))) {
				goto brk;
			}
			user2 += sizeof(in_sess);
			in_len -= sizeof(in_sess);
			DEBUGP("xt_do_get_ctl: in_sess={"
				".source[A]={"
					"RTP={.addr=0x%x:%u, .learning_timeout=%u, .flags=%x},"
					"RTCP={.addr=0x%x:%u, .learning_timeout=%u, .flags=%x},"
					".always_learn=%u},"
				".source[B]={"
					"RTP={.addr=0x%x:%u, .learning_timeout=%u, .flags=%x},"
					"RTCP={.addr=0x%x:%u, .learning_timeout=%u, .flags=%x},"
					".always_learn=%u}"
				"}\n",
				in_sess.source[0].stream[0].addr.ip, in_sess.source[0].stream[0].addr.port, in_sess.source[0].stream[0].learning_timeout, in_sess.source[0].stream[0].flags,
				in_sess.source[0].stream[1].addr.ip, in_sess.source[0].stream[1].addr.port, in_sess.source[0].stream[1].learning_timeout, in_sess.source[0].stream[1].flags,
				in_sess.source[0].always_learn,
				in_sess.source[1].stream[0].addr.ip, in_sess.source[1].stream[0].addr.port, in_sess.source[1].stream[0].learning_timeout, in_sess.source[1].stream[0].flags,
				in_sess.source[1].stream[1].addr.ip, in_sess.source[1].stream[1].addr.port, in_sess.source[1].stream[1].learning_timeout, in_sess.source[1].stream[1].flags,
				in_sess.source[1].always_learn
			);
			spin_lock_bh(&xt_lock);
			for (sess_id = in_sess_id.sess_id_min, sess=p->sessions+in_sess_id.sess_id_min; sess_id <= in_sess_id.sess_id_max && sess_id < p->so.max_sess; sess_id++, sess++) {
				int new_state;
				u_int32_t new_expires;

				xt_do_session_audit(&stamp, p, sess);
				if (sess->state == xt_rtpproxy_EXPIRED || sess->state == xt_rtpproxy_DESTROYED)
					continue; /* we cannot update such sessions */

				if (in_sess_id.created && sess->created != in_sess_id.created)
					continue; /* created does not match */

				new_state = xt_rtpproxy_NONE;
				new_expires = stamp.tv_sec+p->so.expires_timeout;
				DEBUGP("xt_do_get_ctl: sess_id=%d, new_expires=%u, updating\n", sess_id, new_expires);
				for (i=0; i<XT_RTPPROXY_MAX_GATE; i++) {
					int k;
					/* if forced switchboard that was demanded in opposite order we must assign in_sess[0]->sess[1], in_sess[1]->sess[0] */
					k = opposite_flag?XT_RTPPROXY_MAX_GATE-i-1:i;
					for (j=0; j<2; j++) {
						if (in_sess.source[k].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR) {
							if (sess->gate[i].stream[j].source.ip != in_sess.source[k].stream[j].addr.ip ||
							    sess->gate[i].stream[j].source.port != in_sess.source[k].stream[j].addr.port) {
								sess->gate[i].stream[j].source = in_sess.source[k].stream[j].addr;
								new_state = xt_rtpproxy_INIT1;
							}
						}
						else if (sess->state == xt_rtpproxy_NONE)
							memset(&sess->gate[i].stream[j].source, 0, sizeof(sess->gate[i].stream[j].source));

						if (in_sess.source[k].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT) {
							if (in_sess.source[k].stream[j].learning_timeout) {
								sess->gate[i].stream[j].learning = stamp.tv_sec+in_sess.source[k].stream[j].learning_timeout;
								if (sess->gate[i].stream[j].learning > new_expires)
									new_expires = sess->gate[i].stream[j].learning;
								new_state = xt_rtpproxy_INIT1;
							}
							else
								sess->gate[i].stream[j].learning = 0;
						}
						else if (sess->state == xt_rtpproxy_NONE)
							sess->gate[i].stream[j].learning = 0;
					}	
					sess->gate[i].always_learn = in_sess.source[k].always_learn;
					if (sess->state == xt_rtpproxy_NONE)
						memset(&sess->gate[i].stat, 0, sizeof(sess->gate[i].stat));			
				}
				if (new_state == xt_rtpproxy_NONE) {
				/* we do not allow session destroying or do not update when no change made */
					continue;
				}
				sess->expires = new_expires;
				/* maybe both sides initialized ? */
				for (j=0; j<2; j++) {
					for (i=0; i<XT_RTPPROXY_MAX_GATE; i++) {
						if (!sess->gate[i].stream[j].learning && !sess->gate[i].stream[j].source.ip)
							break;
					}
					if (i >= XT_RTPPROXY_MAX_GATE) { /* all source IPs known/learning for at least one stream ? */
						new_state = xt_rtpproxy_INIT2;
						break;
					}
				}
				if (sess->state == xt_rtpproxy_NONE)
					p->so.stat.counter.alloc++;
				else
					p->so.stat.counter.updated++;

				/* we change always to INIT state (even already in FORWARD state) */
				p->so.stat.state[sess->state]--;
				sess->state = new_state;
				p->so.stat.state[sess->state]++;
			}
			spin_unlock_bh(&xt_lock);
			ret = 0;
			break;
		}
		case XT_RTPPROXY_SO_SET_DELETE_SESSION|XT_OPT_SET_CMD: {
			struct xt_rtpproxy_switchboard *p;
			int opposite_flag;
			do_gettimeofday(&stamp);
			p = (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD)?xt_find_switchboard(&in_switchboard, &opposite_flag):switchboard_reg;
			spin_lock_bh(&xt_lock);
			while (p) {
				int sess_id;
				for (sess_id=in_sess_id.sess_id_min; sess_id<=in_sess_id.sess_id_max && sess_id<p->so.max_sess; sess_id++) {
					if (in_sess_id.created && p->sessions[sess_id].created != in_sess_id.created)
						continue; /* created does not match */
					switch (p->sessions[sess_id].state) {
						case xt_rtpproxy_INIT1:
						case xt_rtpproxy_INIT2:
						case xt_rtpproxy_FORWARD1:
						case xt_rtpproxy_FORWARD2:
							DEBUGP("xt_do_get_ctl: sess_id=%d, destroying\n", sess_id);
							p->sessions[sess_id].expires = stamp.tv_sec;  /* required for and packet dropping and ressurection */
							p->so.stat.state[p->sessions[sess_id].state]--;
							p->sessions[sess_id].state = xt_rtpproxy_DESTROYED;
							p->so.stat.state[p->sessions[sess_id].state]++;
							p->so.stat.counter.destroyed++;
							break;
						default: ;
					}
				}
				if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD)
					break;
				p = p->next;
			}
			spin_unlock_bh(&xt_lock);
			ret = 0;
			break;
		}
		default:
			so_global.stat.unknown++;
			PRINTK(KERN_WARNING, "xt_%s_ctl: unknown request %i\n", cmd&XT_OPT_SET_CMD?"set":"get", cmd&~XT_OPT_SET_CMD);
			ret = -EINVAL;
			goto ret;
	}
	if (in_header.flags & XT_RTPPROXY_SOCKOPT_FLAG_IN_RESET_GLOBAL_STAT) {
		memset(&so_global.stat, 0, sizeof(so_global.stat));
	}
brk:
	if (cmd & XT_OPT_SET_CMD)
		so_global.stat.cmd[(cmd&~XT_OPT_SET_CMD)-XT_RTPPROXY_SO_BASE_CTL]++;
	else
		so_global.stat.cmd[cmd-XT_RTPPROXY_SO_BASE_CTL+XT_RTPPROXY_SO_SET_MAX-XT_RTPPROXY_SO_BASE_CTL+1]++;
fault:	
	if (ret == 0)
		so_global.stat.ok++;
	else
		so_global.stat.fail++;
ret:
	up(&xt_usercontext_mutex);

ret_no_unlock:
	DEBUGP("xt_do_get_ctl: ret=%d\n", ret);
	return ret;
}

static int xt_do_set_ctl(struct sock *sk, int cmd,
                               void *user, u_int32_t len)
{
	return xt_do_get_ctl(sk, cmd|XT_OPT_SET_CMD, user, (int *) &len);
}

/* user-context */
static 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
	int 
#else
	bool
#endif
		do_checkentry(const char *tablename,
				const void *e_void,
				const struct xt_target *target,
				void *targetinfo,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
				unsigned int targinfosize,
#endif
				unsigned int hook_mask)
{
	/*const*/ struct xt_rtpproxy_target_info *info = targetinfo;
	const struct ipt_entry *e = e_void;
	int j;
	unsigned l1, l2;

	DEBUGP("do_checkentry(tablename='%s', e=%p, targetinfo=%p, hook_mask=%x)\n", tablename, e, targetinfo, hook_mask);

#if defined(CONFIG_NF_CONNTRACK) || defined(CONFIG_NF_CONNTRACK_MODULE)
	if (nf_ct_l3proto_try_module_get(target->family) < 0) {
		PRINTK(KERN_WARNING, "do_checkentry(): can't load conntrack support for proto=%d\n", target->family);
		return 0;
	}
#endif
		
	if (strcmp(tablename, "mangle") != 0) {
		PRINTK(KERN_WARNING, "bad table '%s', use the 'mangle' table.\n",
			   tablename);
		return 0;
	}
	if (hook_mask & ~(  (1 << NF_IP_PRE_ROUTING)
			    | (1 << NF_IP_LOCAL_OUT)
			    | (1 << NF_IP_POST_ROUTING)
			    | (1 << NF_IP_LOCAL_IN)) ) {
		PRINTK(KERN_WARNING, "do_checkentry(): bad hook_mask=%x\n", hook_mask);
		return 0;
	}

	/* Must specify that it's a UDP packet */
	if (e->ip.proto != IPPROTO_IP) {  /* if specified a protocol parameter */
		if (e->ip.proto != IPPROTO_UDP || (e->ip.invflags & XT_INV_PROTO)) {
			PRINTK(KERN_WARNING, "not valid for non-udp (proto:%c0x%x)\n",(e->ip.invflags & XT_INV_PROTO)?'!':' ', e->ip.proto);
			return 0;
		}
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18)
	if (targinfosize != XT_ALIGN(sizeof(*info))) {
		PRINTK(KERN_WARNING, "targinfosize %u != %u\n",
		       targinfosize,
		       XT_ALIGN(sizeof(struct xt_rtpproxy_target_info)));
		return 0;
	}
#endif
	l1 = info->gate[0].port + 2*info->max_sess;
	l2 = info->gate[OPPOSITE_GATE(0)].port + 2*info->max_sess;
	if ( l1 > 0xFFFFL) {
		PRINTK(KERN_WARNING, "end of port-a is out of range\n");
		return 0;
	}
	if ( !info->gate[OPPOSITE_GATE(0)].port || l2 > 0xFFFFL) {
		PRINTK(KERN_WARNING, "end of port-b is out of range\n");
		return 0;
	}
	if (info->gate[0].ip == info->gate[OPPOSITE_GATE(0)].ip) {
		if ( (l2 > info->gate[0].port && info->gate[OPPOSITE_GATE(0)].port < l1) ||
		     (l1 > info->gate[OPPOSITE_GATE(0)].port && info->gate[0].port < l2) ) {
			PRINTK(KERN_WARNING, "port-a and port-b ranges are overlapped\n");
			return 0;
		}
	}

	if (down_interruptible(&xt_usercontext_mutex)) {
		return -EINTR;
	}
	/* check if session is already registered */
	for (info->switchboard = switchboard_reg; info->switchboard; info->switchboard = info->switchboard->next ) {
		if (info->switchboard->so.max_sess == info->max_sess) {
			for (j=0; j<XT_RTPPROXY_MAX_GATE; j++) {
				if ( info->switchboard->so.gate[0].addr.ip == info->gate[j].ip &&
				     info->switchboard->so.gate[0].addr.port == info->gate[j].port &&
				     info->switchboard->so.gate[OPPOSITE_GATE(0)].addr.ip == info->gate[OPPOSITE_GATE(j)].ip &&
				     info->switchboard->so.gate[OPPOSITE_GATE(0)].addr.port == info->gate[OPPOSITE_GATE(j)].port ) {
					DEBUGP("do_checkentry(): found switchboard=%p\n", info->switchboard);
					goto found;
				}
			}
		}
	}

	DEBUGP("do_checkentry(): allocating new switchboard\n");
	DEBUGP("do_checkentry(): vmalloc(%u)\n", sizeof(*info->switchboard));
	info->switchboard = vmalloc(sizeof(*info->switchboard)); /* alternatively kmalloc(, GPF_KERNEL); */
	if (!info->switchboard) {
		PRINTK(KERN_WARNING, "cannot allocate %u bytes\n",
			sizeof(*info->switchboard));
		return 0;
	}
	DEBUGP("do_checkentry(): allocated info->switchboard=%p\n", info->switchboard);
	memset(info->switchboard, 0, sizeof(*(info->switchboard)));
	info->switchboard->so.max_sess = info->max_sess;
	info->switchboard->so.resurrection_timeout = info->resurrection_timeout;
	info->switchboard->so.expires_timeout = info->expires_timeout;
	info->switchboard->so.stat.state[xt_rtpproxy_NONE] = info->max_sess;
	for (j=0; j<XT_RTPPROXY_MAX_GATE; j++) {
		info->switchboard->so.gate[j].addr = info->gate[j];
	}
	DEBUGP("do_checkentry(): vmalloc(%u)\n", sizeof(*info->switchboard->sessions)*info->max_sess);
	info->switchboard->sessions = vmalloc(sizeof(*info->switchboard->sessions)*info->max_sess);
	if (!info->switchboard->sessions) {
		PRINTK(KERN_WARNING, "cannot allocate %u bytes\n",
			sizeof(*info->switchboard->sessions)*info->max_sess);
		vfree(info->switchboard);
		return 0;
	}
	DEBUGP("do_checkentry(): allocated info->switchboard->sessions=%p\n", info->switchboard->sessions);
	memset(info->switchboard->sessions, 0, sizeof(*info->switchboard->sessions)*info->max_sess);
	for (j=0; j<info->max_sess; j++) {
		int i;
		if (j > 0) {
			info->switchboard->sessions[j-1].next = info->switchboard->sessions + j;
			info->switchboard->sessions[j].prev = info->switchboard->sessions + j-1;
		}
			
		info->switchboard->sessions[j].sess_id = j;
		info->switchboard->sessions[j].state = xt_rtpproxy_NONE;
		for (i=0; i<XT_RTPPROXY_MAX_GATE; i++) {
			info->switchboard->sessions[j].gate[i].stream[0].port = info->gate[i].port+2*j;
			info->switchboard->sessions[j].gate[i].stream[1].port = info->gate[i].port+2*j+1;
		}
	}
	/* add self at the top of registration chain */
	spin_lock_bh(&xt_lock);  /* write lock for switchboard_reg chain to see it interrupt-context consistently */
	info->switchboard->next = switchboard_reg;
	if (switchboard_reg) {
		switchboard_reg->prev = info->switchboard;
	}
	switchboard_reg = info->switchboard;
	spin_unlock_bh(&xt_lock);
	so_global.total.num_switchboard++;
	so_global.total.num_session += info->switchboard->so.max_sess;
		
found:
	info->switchboard->usage++;
	DEBUGP("do_checkentry(): info->switchboard->usage=%d\n", info->switchboard->usage);
	up(&xt_usercontext_mutex);
	return 1;
}

/* user-context */
static void do_destroy(
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
                    const struct xt_target *target,
#endif
                    void *targetinfo
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
					, unsigned int targetsize
#endif
					)
{
	struct xt_rtpproxy_target_info *info = targetinfo;
	DEBUGP("destroy(targetinfo=%p)\n", targetinfo);
	down(&xt_usercontext_mutex);
	spin_lock_bh(&xt_lock); /* write lock for switchboard_reg chain to see it interrupt-context consistently */
	info->switchboard->usage--;
	if (!info->switchboard->usage) {
		if (info->switchboard->prev) {
			info->switchboard->prev->next = info->switchboard->next;
		}
		else
			switchboard_reg = info->switchboard->next;
		if (info->switchboard->next)
			info->switchboard->next->prev = info->switchboard->prev;
		so_global.total.num_switchboard--;
		so_global.total.num_session -= info->switchboard->so.max_sess;		
	}
	else {
		DEBUGP("destroy(): info->switchboard->usage=%d\n", info->switchboard->usage);
		info = NULL;  /* do not free switchboard */
	}
	up(&xt_usercontext_mutex);
	spin_unlock_bh(&xt_lock);

	if (info) {
		DEBUGP("destroy(): vfree(info->switchboard->sessions=%p)\n", info->switchboard->sessions);
		vfree(info->switchboard->sessions);
		DEBUGP("destroy(): Deallocated %u bytes\n",
			sizeof(*info->switchboard->sessions)*info->max_sess);

		DEBUGP("destroy(): vfree(info->switchboard=%p)\n", info->switchboard);
		vfree(info->switchboard);
	        DEBUGP("destroy(): Deallocated %u bytes\n",
			sizeof(*info->switchboard));
	}
}


static struct xt_target xt_rtpproxy_reg = {
	.name = "RTPPROXY",
	.family = AF_INET,
	.proto = IPPROTO_UDP,
	.target = do_target,
	.targetsize = sizeof(struct xt_rtpproxy_target_info),
	.checkentry = do_checkentry,
	.destroy = do_destroy,
	.hooks =  (1 << NF_IP_PRE_ROUTING) | (1 << NF_IP_POST_ROUTING) | (1 << NF_IP_LOCAL_OUT) | (1 << NF_IP_LOCAL_IN),
	.table = "mangle",
	.me = THIS_MODULE
};

static struct nf_sockopt_ops xt_rtpproxy_sockopts = {
	.pf = PF_INET,
	.set_optmin = XT_RTPPROXY_SO_BASE_CTL,
	.set_optmax = XT_RTPPROXY_SO_SET_MAX+1,
	.set = xt_do_set_ctl,
	.get_optmin = XT_RTPPROXY_SO_BASE_CTL,
	.get_optmax = XT_RTPPROXY_SO_GET_MAX+1,
	.get = xt_do_get_ctl
};


static int __init init(void)
{
	struct timeval stamp;
	PRINTK(KERN_INFO, "initializing\n");

	switchboard_reg = NULL;
	memset(&so_global, 0, sizeof(so_global));
	so_global.h.version = XT_RTPPROXY_SOCKOPT_VERSION;
	do_gettimeofday(&stamp);
	so_global.started = stamp.tv_sec;

	/* Register setsockopt */
	DEBUGP("init(): nf_register_sockopt(%p)\n", &xt_rtpproxy_sockopts);
	if (nf_register_sockopt(&xt_rtpproxy_sockopts) < 0) {
		PRINTK(KERN_ERR, "Can't register sockopts. Aborting\n");
		goto error_cleanup;
	}

	DEBUGP("init(): xt_register_target(%p)\n", &xt_rtpproxy_reg);
	if (xt_register_target(&xt_rtpproxy_reg)) {
		PRINTK(KERN_ERR, "Can't register target. Aborting\n");
		goto error_cleanup2;				
	}
	return 0;

error_cleanup2:
	DEBUGP("init(): nf_unregister_sockopt(%p)\n", &xt_rtpproxy_sockopts);
	nf_unregister_sockopt(&xt_rtpproxy_sockopts);
error_cleanup:
	return -EINVAL;
}


static void __exit fini(void)
{
	PRINTK(KERN_INFO, "finalizing\n");
	DEBUGP("fini(): xt_unregister_target(%p)\n", &xt_rtpproxy_reg);
	xt_unregister_target(&xt_rtpproxy_reg);
	DEBUGP("fini(): nf_unregister_sockopt(%p)\n", &xt_rtpproxy_sockopts);
	nf_unregister_sockopt(&xt_rtpproxy_sockopts);
}

module_init(init);
module_exit(fini);

